//		Copyright © 1997–2000  Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//		Client monitor class for the Csound Server.

#include <cstdio>

#include <Alert.h>
#include <InterfaceDefs.h>
#include <OS.h>
#include <Point.h>
#include <Rect.h>
#include <SupportDefs.h>

#include "CsSvInterface.h"
#include "CsSvClient.h"
#include "CsSvGraphicsView.h"
#include "CsSvInputView.h"
#include "CsSvStringReader.h"
#include "CsSvWindow.h"

// Constant data.
static const char kClientThreadName[] = "Csound client thread";
static const char kInputWindowName[] = "X/Y Input";

// Static data members.
int32 CsoundServerClient::smClientCounter = 0;

CsoundServerClient::CsoundServerClient(ServerClientPort *pClientPort)
:	mId((uint32)atomic_add(&smClientCounter, 1)),
	mRequestPort(pClientPort->mRequestPort),
	mReplyPort(pClientPort->mReplyPort),
	mClientThread(B_ERROR),
	mClientSem(B_ERROR)
{
	// empty
}

CsoundServerClient::~CsoundServerClient(void)
{
	if (mClientSem >= B_OK) {
		acquire_sem(mClientSem);

		kill_thread(mClientThread);
		mClientThread = B_ERROR;

		sem_id sem = mClientSem;
		mClientSem = B_ERROR;
		delete_sem(sem);
	}
}

status_t
CsoundServerClient::Run(void)
{
	status_t status = B_ERROR;
	
	if (mRequestPort >= B_OK) {
		// Create the semaphore.
		mClientSem = create_sem(1, "Client thread lock");
		if (mClientSem >= B_OK) {
			// Spawn the client thread.
			mClientThread =
				spawn_thread(ClientThreadStub, kClientThreadName, B_NORMAL_PRIORITY, this);
			if (mClientThread >= B_OK) {
				// Start it running.
				if (resume_thread(mClientThread) < B_OK) {
					// Dang.
					kill_thread(mClientThread);
					mClientThread = B_ERROR;

				} else {
					status = B_OK;
				}
			}
			
			if (mClientThread < B_OK) {
				// Drat.
				delete_sem(mClientSem);
				mClientSem = B_ERROR;
			}
		}
	}
	
	return status;
}

// Thread functions.
int32
CsoundServerClient::ClientThreadStub(void *pData)
{
	return ((CsoundServerClient *)pData)->ClientThread();
}

int32
CsoundServerClient::ClientThread(void)
{
	// Read messages from the client port and act on them.
	bool finished = false;
	while (!finished) {
		int32 msgCode;
		char msgBuffer[ServerBufferSize];
		if (read_port(mRequestPort, &msgCode, msgBuffer, sizeof(msgBuffer))
			 >= B_OK) {
			switch (msgCode) {
			
			case kServerNewGraphWindow:
				// Open a new graphics window.
				NewGraphWindow((ServerGraphWindow *)msgBuffer);
				break;
			
			case kServerDrawGraph:
				// Draw a graph.
				DrawGraph((ServerGraphData *)msgBuffer);
				break;
			
			case kServerNewInputWindow:
				// Open a new X/Y input window.
				NewInputWindow((ServerInputWindow *)msgBuffer);
				break;
			
			case kServerReadInput:
				// Read a position from an input window.
				ReadInput((ServerInputRequest *)msgBuffer);
				break;
			
			case kServerKillWindow:
				// Get rid of a window.
				KillWindow((ServerKillWindow *)msgBuffer);
				break;
			
			default:
				// Sounds foreign to me.
				sprintf(msgBuffer, "Csound Server: Client 0x%08lx Huh?", msgCode);
				(new BAlert("", msgBuffer, "Duh" B_UTF8_ELLIPSIS))->Go();
				break;
			}
		} else {
			// Client has closed connection.
			finished = true;
		}
	}
	
	// Clean up.
	acquire_sem(mClientSem);
	mClientThread = B_ERROR;
	sem_id sem = mClientSem;
	mClientSem = B_ERROR;
	delete_sem(sem);

	return 0;
}

void
CsoundServerClient::NewGraphWindow(ServerGraphWindow *pDesc)
{
	// Read the window name from the port.
	CsoundServerStringReader nameReader(mRequestPort);
	
	// Create the graphics view, add it to a new window, and show the window.
	const BRect frame(
		pDesc->mX, pDesc->mY, pDesc->mX + pDesc->mW - 1, pDesc->mY + pDesc->mH - 1);
	CsoundServerGraphicsView *pView =
		new CsoundServerGraphicsView(frame, nameReader);
	const CsoundServerWindow::ID id(mId, pDesc->mSerialNo);
	CsoundServerWindow *pWindow = new CsoundServerWindow(frame, nameReader, id);
	pWindow->AddChild(pView);
	pWindow->Show();
}

void
CsoundServerClient::DrawGraph(ServerGraphData *pDesc)
{
	// Read the coordinates and caption from the port.
	const uint32 nAxesCoords = 2 * pDesc->mAxesLineCount * 2;
	float *const pAxesPoints = new float[nAxesCoords];
	const uint32 nGraphCoords = (pDesc->mGraphLineCount+1) * 2;
	float *const pGraphPoints = new float[nGraphCoords];
	int32 dummy;
	status_t status =
		read_port(mRequestPort, &dummy, pAxesPoints, nAxesCoords * sizeof(float));
	if (status >= B_OK) {
		status =
			read_port(mRequestPort, &dummy, pGraphPoints, nGraphCoords*sizeof(float));
	}
	CsoundServerStringReader captionReader(mRequestPort);
	if (status >= B_OK && (const char *)captionReader == 0) {
		status = B_ERROR;
	}

	if (status >= B_OK) {
		// Find appropriate window and view.
		const CsoundServerWindow::ID id(mId, pDesc->mSerialNo);
		CsoundServerWindow *const pWindow = CsoundServerWindow::Find(id);
		if (pWindow->Lock()) {
			CsoundServerGraphicsView *const pView =
				dynamic_cast<CsoundServerGraphicsView *>(pWindow->ChildAt(0));
			if (pView) {
				pView->AddPage(pDesc, pAxesPoints, pGraphPoints, captionReader);
			}
			pWindow->Unlock();
		}
	}
	
	delete [] pGraphPoints;
	delete [] pAxesPoints;
}

void
CsoundServerClient::NewInputWindow(ServerInputWindow *pDesc)
{
	// Create the input view, add it to a new window, and show the window.
	const BRect frame(
		pDesc->mX, pDesc->mY, pDesc->mX + pDesc->mW - 1, pDesc->mY + pDesc->mH - 1);
	CsoundServerInputView *pView =
		new CsoundServerInputView(frame, kInputWindowName,
										  pDesc->mValueX,
										  pDesc->mValueY);
	const CsoundServerWindow::ID id(mId, pDesc->mSerialNo);
	CsoundServerWindow *pWindow =
		new CsoundServerWindow(frame, kInputWindowName, id);
	pWindow->AddChild(pView);
	pWindow->Show();
}

void
CsoundServerClient::ReadInput(ServerInputRequest *pDesc)
{
	ServerInputReply reply = { 0.0, 0.0, 0 };
	
	// Find appropriate window and view.
	const CsoundServerWindow::ID id(mId, pDesc->mSerialNo);
	CsoundServerWindow *const pWindow = CsoundServerWindow::Find(id);
	if (pWindow->Lock()) {
		CsoundServerInputView *const pView =
			dynamic_cast<CsoundServerInputView *>(pWindow->ChildAt(0));
		if (pView) {
			pView->GetState(reply.mValueX, reply.mValueY, reply.mDown);
		}
		pWindow->Unlock();
	}
	
	// Send reply to client.
	write_port(mReplyPort, 0, &reply, sizeof(reply));
}

void
CsoundServerClient::KillWindow(ServerKillWindow *pDesc)
{
	// Find appropriate window and close it.
	const CsoundServerWindow::ID id(mId, pDesc->mSerialNo);
	CsoundServerWindow *const pWindow = CsoundServerWindow::Find(id);
	if (pWindow->Lock()) {
		if (pWindow->QuitRequested()) {
			pWindow->Quit();
		} else {
			pWindow->Unlock();
		}
	}
}
