//		Copyright © 1997	Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//		Graphics view class for the Csound Server.

#include <cstring>

#include <Debug.h>
#include <Font.h>
#include <List.h>
#include <OS.h>
#include <Point.h>
#include <Rect.h>

#include "CsSvGraphicsView.h"
#include "CsSvInterface.h"

CsoundServerGraphicsView::CsoundServerGraphicsView(
	BRect windowFrame, const char *pWindowName)
:	inherited(windowFrame, pWindowName),
	mPages(),
	mCurrentPage(0)
{
	// empty
}

CsoundServerGraphicsView::~CsoundServerGraphicsView(void)
{
	void *pItem;
	for (uint32 i = 0; (pItem = mPages.ItemAt(i)) != 0; ++i) {
		delete (CsoundServerGraphicsPage *)pItem;
	}
	mPages.MakeEmpty();
}

void
CsoundServerGraphicsView::Draw(BRect)
{
	BFont font;
	GetFont(&font);
	font_height height;
	font.GetHeight(&height);

	CsoundServerGraphicsPage *pPage =
		(CsoundServerGraphicsPage *)mPages.ItemAt(mCurrentPage);
	if (pPage) {
		pPage->DrawOn(this);

		const rgb_color textColor = { 255, 255, 255, 0 };
		SetHighColor(textColor);
		DrawString(pPage->Caption(), BPoint(0, height.ascent));
		
		const uint32 nPages = mPages.CountItems();
		if (nPages > 1) {
			char buffer[100];
			sprintf(buffer, " -- Page %ld of %ld", mCurrentPage + 1, nPages);
			DrawString(buffer);
		}
	} else {
		DrawString("No data available", BPoint(0, height.ascent));
	}
	Flush();
}

void
CsoundServerGraphicsView::KeyDown(const char *pBytes, int32 nBytes)
{
	if (nBytes == 1) {
		switch (*pBytes) {
		
		// Go forward.
		case B_ENTER:
		case B_SPACE:
		case B_TAB:
		case B_RIGHT_ARROW:
		case B_DOWN_ARROW:
		case B_PAGE_DOWN:
		case '+':
			if (!mPages.ItemAt(++mCurrentPage)) {
				mCurrentPage = 0;
			}
			Invalidate();
			return;
		
		// Go backward.
		case B_BACKSPACE:
		case B_LEFT_ARROW:
		case B_UP_ARROW:
		case B_PAGE_UP:
		case '-':
			if (!mPages.ItemAt(--mCurrentPage)) {
				mCurrentPage = mPages.CountItems() - 1;
			}
			Invalidate();
			return;
		
		// Go to first/last page.
		case B_HOME:
			mCurrentPage = 0;
			Invalidate();
			return;
		case B_END:
			mCurrentPage = mPages.CountItems() - 1;
			Invalidate();
			return;
		
		default:
			inherited::KeyDown(pBytes, nBytes);
			break;
		}
	}
}

void
CsoundServerGraphicsView::AddPage(
	const ServerGraphData *pDesc,
	const float *pAxesPoints, const float *pGraphPoints, const char *pCaption)
{
	if (LockLooper()) {
		CsoundServerGraphicsPage *pNewPage =
			new CsoundServerGraphicsPage(pDesc,
												  pAxesPoints, pGraphPoints, pCaption);
		mPages.AddItem(pNewPage);
		mCurrentPage = mPages.CountItems() - 1;
	
		// Force refresh.	
		Invalidate();
	
		UnlockLooper();
	}
}

CsoundServerGraphicsView::CsoundServerGraphicsPage::CsoundServerGraphicsPage(
	const ServerGraphData *pDesc,
	const float *pAxesPoints, const float *pGraphPoints, const char *pCaption)
:	mAxesLineCount(pDesc->mAxesLineCount),
	mpAxesPoints(new BPoint[2 * mAxesLineCount]),
	mGraphLineCount(pDesc->mGraphLineCount),
	mpGraphPoints(new BPoint[mGraphLineCount + 1]),
	mpCaption(strcpy(new char[strlen(pCaption)+1], pCaption)),
	mMidAxis(pDesc->mMidAxis)
{
	// Axes.
	for (uint32 i = 0; i < 2 * mAxesLineCount; ++i) {
		const float x = *pAxesPoints++;
		const float y = *pAxesPoints++;
		mpAxesPoints[i] = BPoint(x, y);
	}
	
	// Graph.
	for (uint32 i = 0; i <= mGraphLineCount; ++i) {
		const float x = *pGraphPoints++;
		const float y = *pGraphPoints++;
		mpGraphPoints[i] = BPoint(x, y);
	}
}

CsoundServerGraphicsView::CsoundServerGraphicsPage::~CsoundServerGraphicsPage(void)
{
	delete [] mpAxesPoints;
	delete [] mpGraphPoints;
	delete [] mpCaption;
}

void
CsoundServerGraphicsView::CsoundServerGraphicsPage::DrawOn(BView *pView) const
{
	const rgb_color axesColor = { 255,   0,   0, 0 };
	const rgb_color grphColor = {   0, 255,   0, 0 };
	
	BRect bounds = pView->Bounds();
	const float xScale = bounds.Width();
	const float yScale = bounds.Height();
	
	pView->SetHighColor(axesColor);
	pView->StrokeLine(BPoint(xScale * mMidAxis, 0.0),
							BPoint(xScale * mMidAxis, yScale),
							B_MIXED_COLORS);
	
	if (mAxesLineCount) {
		pView->BeginLineArray(mAxesLineCount);
		uint32 pnt = 0;
		for (uint32 i = 0; i < mAxesLineCount; ++i) {
			BPoint start = mpAxesPoints[pnt++];
			start.x *= xScale;
			start.y *= yScale;
			BPoint end   = mpAxesPoints[pnt++];
			end.x *= xScale;
			end.y *= yScale;
			pView->AddLine(start, end, axesColor);
		}
		pView->EndLineArray();
	}

	if (mGraphLineCount) {
		pView->BeginLineArray(mGraphLineCount);
		BPoint start = mpGraphPoints[0];
		start.x *= xScale;
		start.y *= yScale;
		for (uint32 i = 1; i <= mGraphLineCount; ++i) {
			BPoint end = mpGraphPoints[i];
			end.x *= xScale;
			end.y *= yScale;
			pView->AddLine(start, end, grphColor);
			start = end;
		}
		pView->EndLineArray();
	}
}
