//		Copyright © 2000  Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//		MIDI producer class for the Csound Server.

#include "CsSvMidiProducer.h"

#include <cstring>

#include <Autolock.h>
#include <List.h>
#include <Locker.h>
#include <MidiProducer.h>
#include <OS.h>

#include "CsSvInterface.h"

namespace
{
	const char kProducerThreadName[] = "Csound MIDI output";
}

BList CsoundServerMidiProducer::smProducerList;
BLocker CsoundServerMidiProducer::smProducerListLocker("producer list");

CsoundServerMidiProducer::CsoundServerMidiProducer(
	const char *pName, const ServerMidiPort *pPort)
:	inherited(pName),
	mPort(pPort->mPort),
	mpBuffer(0),
	mBufferSize(0)
{
	// Register the instance locally.
	BAutolock lock(smProducerListLocker);
	smProducerList.AddItem(this);
}

CsoundServerMidiProducer::~CsoundServerMidiProducer(void)
{
	delete [] mpBuffer;
	mpBuffer = 0;
	mBufferSize = 0;
}

status_t
CsoundServerMidiProducer::Run(void)
{
	// Start a thread which reads data from the client.
	thread_id producerThread =
		spawn_thread(ProducerThreadStub, kProducerThreadName, B_NORMAL_PRIORITY, this);
	if (producerThread >= B_OK) {
		// Start it running.
		status_t status = resume_thread(producerThread);
		if (status < B_OK) {
			// Dang.
			kill_thread(producerThread);
			return status;
		}
	}
	return (status_t)producerThread;
}

// Find an active producer (needed to work around a bug in BMidiRoster).
CsoundServerMidiProducer *
CsoundServerMidiProducer::Find(const char *pName)
{
	BAutolock lock(smProducerListLocker);

	const int32 nItems = smProducerList.CountItems();
	for (int32 i = 0; i < nItems; ++i) {
		CsoundServerMidiProducer *pProducer =
			static_cast<CsoundServerMidiProducer *>(smProducerList.ItemAt(i));
		if (pProducer && !strcmp(pName, pProducer->Name())) {
			pProducer->Acquire();
			return pProducer;
		}
	}
	
	return 0;
}

// Thread functions.
int32
CsoundServerMidiProducer::ProducerThreadStub(void *pData)
{
	return ((CsoundServerMidiProducer *)pData)->ProducerThread();
}

int32
CsoundServerMidiProducer::ProducerThread(void)
{
	// Read data from the client until it closes the port (in which case
	// we'll get an error).
	ssize_t bufferSize;
	while ((bufferSize = port_buffer_size(mPort)) >= 0) {
		if (bufferSize > mBufferSize) {
			delete [] mpBuffer;
			mpBuffer = new char[mBufferSize = bufferSize];
		}
		
		int32 dummy;
		if ((bufferSize = read_port(mPort, &dummy, mpBuffer, mBufferSize)) < B_OK) {
			break;
		}
		
		// Send data to consumers.
		SprayData(mpBuffer, bufferSize, true, system_time());
	}
	
	// Unregister and release the MIDI producer.
	Unregister();
	{
		// Unregister the instance locally.
		BAutolock lock(smProducerListLocker);
		smProducerList.RemoveItem(this);
	}
	Release();

	// Done.	
	return 0;
}
