//		Copyright © 1997–2000  Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//		Audio conversion routines (taken from the SOX package).

/*
** This routine converts from linear to ulaw.
**
** Craig Reese: IDA/Supercomputing Research Center
** Joe Campbell: Department of Defense
** 29 September 1989
**
** References:
** 1) CCITT Recommendation G.711  (very difficult to follow)
** 2) "A New Digital Technique for Implementation of Any
**    Continuous PCM Companding Law," Villeret, Michel,
**    et al. 1973 IEEE Int. Conf. on Communications, Vol 1,
**    1973, pg. 11.12-11.17
** 3) MIL-STD-188-113,"Interoperability and Performance Standards
**    for Analog-to_Digital Conversion Techniques,"
**    17 February 1987
**
** Input: Signed 16 bit linear sample
** Output: 8 bit ulaw sample
*/

#undef ZEROTRAP      /* turn off the trap as per the MIL-STD */
#define uBIAS 0x84   /* define the add-in bias for 16 bit samples */
#define uCLIP 32635
#define ACLIP 31744

unsigned char
st_linear_to_ulaw(int sample)
{
	static int exp_lut[256] = {0,0,1,1,2,2,2,2,3,3,3,3,3,3,3,3,
									   4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,
									   5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
									   5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
									   6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
									   6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
									   6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
									   6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
									   7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
									   7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
									   7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
									   7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
									   7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
									   7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
									   7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
									   7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7};
	int sign, exponent, mantissa;
	unsigned char ulawbyte;

	/* Get the sample into sign-magnitude. */
	sign = (sample >> 8) & 0x80;				/* set aside the sign */
	if (sign != 0) sample = -sample;			/* get magnitude */
	if (sample > uCLIP) sample = uCLIP;		/* clip the magnitude */

	/* Convert from 16 bit linear to ulaw. */
	sample = sample + uBIAS;
	exponent = exp_lut[( sample >> 7 ) & 0xFF];
	mantissa = (sample >> (exponent + 3)) & 0x0F;
	ulawbyte = ~(sign | (exponent << 4) | mantissa);
#ifdef ZEROTRAP
	if (ulawbyte == 0) ulawbyte = 0x02;		/* optional CCITT trap */
#endif

	return ulawbyte;
}

/*
** This routine converts from ulaw to 16 bit linear.
**
** Craig Reese: IDA/Supercomputing Research Center
** 29 September 1989
**
** References:
** 1) CCITT Recommendation G.711  (very difficult to follow)
** 2) MIL-STD-188-113,"Interoperability and Performance Standards
**     for Analog-to_Digital Conversion Techniques,"
**     17 February 1987
**
** Input: 8 bit ulaw sample
** Output: signed 16 bit linear sample
*/

int
st_ulaw_to_linear(unsigned char ulawbyte)
{
	static int exp_lut[8] = { 0, 132, 396, 924, 1980, 4092, 8316, 16764 };
	int sign, exponent, mantissa, sample;

	ulawbyte = ~ulawbyte;
	sign = (ulawbyte & 0x80);
	exponent = (ulawbyte >> 4) & 0x07;
	mantissa = ulawbyte & 0x0F;
	sample = exp_lut[exponent] + (mantissa << (exponent + 3));
	if (sign != 0) sample = -sample;

	return sample;
}

/*
 * A-law routines by Graeme W. Gill.
 * Date: 93/5/7
 *
 * References:
 * 1) CCITT Recommendation G.711
 */

#define ACLIP 31744

unsigned char
st_linear_to_alaw(int sample)
{
	static int exp_lut[128] = {1,1,2,2,3,3,3,3,
									   4,4,4,4,4,4,4,4,
									   5,5,5,5,5,5,5,5,
									   5,5,5,5,5,5,5,5,
									   6,6,6,6,6,6,6,6,
									   6,6,6,6,6,6,6,6,
									   6,6,6,6,6,6,6,6,
									   6,6,6,6,6,6,6,6,
									   7,7,7,7,7,7,7,7,
									   7,7,7,7,7,7,7,7,
									   7,7,7,7,7,7,7,7,
									   7,7,7,7,7,7,7,7,
									   7,7,7,7,7,7,7,7,
									   7,7,7,7,7,7,7,7,
									   7,7,7,7,7,7,7,7,
									   7,7,7,7,7,7,7,7};
	int sign, exponent, mantissa;
	unsigned char alawbyte;

	/* Get the sample into sign-magnitude. */
	sign = ((~sample) >> 8) & 0x80;		/* set aside the sign */
	if (sign == 0) sample = -sample;		/* get magnitude */
	if (sample > ACLIP) sample = ACLIP;	/* clip the magnitude */

	/* Convert from 16 bit linear to alaw. */
	if (sample >= 256) {
		exponent = exp_lut[(sample >> 8) & 0x7F];
		mantissa = (sample >> (exponent + 3)) & 0x0F;
		alawbyte = ((exponent << 4) | mantissa);
	} else
		alawbyte = (sample >> 4);
	alawbyte ^= (sign ^ 0x55);

	return alawbyte;
}

int
st_alaw_to_linear(unsigned char alawbyte)
{
	static int exp_lut[8] = { 0, 264, 528, 1056, 2112, 4224, 8448, 16896 };
	int sign, exponent, mantissa, sample;

	alawbyte ^= 0x55;
	sign = (alawbyte & 0x80);
	alawbyte &= 0x7f;			/* get magnitude */
	if (alawbyte >= 16) {
		exponent = (alawbyte >> 4 ) & 0x07;
		mantissa = alawbyte & 0x0F;
		sample = exp_lut[exponent] + (mantissa << (exponent + 3));
	} else
		sample = (alawbyte << 4) + 8;
	if ( sign == 0 ) sample = -sample;

	return sample;
}
