//		Copyright © 1997	Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//		Describes a Csound fourier (CONVOLVE) analysis job.

#include <cstdio>
#include <cstring>

#include <Message.h>
#include <Path.h>
#include <Point.h>
#include <SupportDefs.h>

#include "CsAbstractJob.h"
#include "CsCVJob.h"
#include "CsCVMainWindow.h"
#include "NaN.h"

// Class name.
static const char kClassName[] = "CsoundCVAnalysisJob";
static const char kJobClass[] = "CV Analysis";

// Window title prefix.
static const char kWindowTitleFormat[] = "Csound CV Analysis: %s";

// Basic C++ object requirements: construct, copy, assign, destroy.
CsoundCVAnalysisJob::CsoundCVAnalysisJob(
	bool assignName, uint32 numFilenames, uint32 numNumbers)
:	inherited(assignName, numFilenames, numNumbers)
{
	// empty
}	

CsoundCVAnalysisJob::CsoundCVAnalysisJob(const CsoundCVAnalysisJob &job)
:	inherited(job)
{
	// empty
}

CsoundCVAnalysisJob &
CsoundCVAnalysisJob::operator =(const CsoundCVAnalysisJob &job)
{
	if (&job != this) {
		inherited::operator =(job);
	}
	return *this;
}

CsoundCVAnalysisJob::~CsoundCVAnalysisJob(void)
{
	// empty
}

CsoundAbstractJob *
CsoundCVAnalysisJob::New(void) const
{
	// Virtual default constructor.
	return new CsoundCVAnalysisJob();
}

CsoundAbstractJob *
CsoundCVAnalysisJob::Clone(void) const
{
	// Virtual copy constructor.
	return new CsoundCVAnalysisJob(*this);
}

CsoundAbstractJob &
CsoundCVAnalysisJob::Become(const CsoundAbstractJob &job)
{
	// Virtual assignment operator.
	return *this = dynamic_cast<const CsoundCVAnalysisJob &>(job);
}

bool
CsoundCVAnalysisJob::operator ==(const CsoundAbstractJob &abstractJob) const
{
	const CsoundCVAnalysisJob *const pJob =
		dynamic_cast<const CsoundCVAnalysisJob *>(&abstractJob);
	if (!pJob) {
		return false;
	}
	if (pJob == this) {
		return true;
	}
	return inherited::operator ==(abstractJob);
}

//	BArchivable requirements.
CsoundCVAnalysisJob::CsoundCVAnalysisJob(BMessage *pMsg)
:	inherited(pMsg)
{
	// empty
}

status_t
CsoundCVAnalysisJob::Archive(BMessage *pMsg, bool deep) const
{
	status_t status = inherited::Archive(pMsg, deep);

	if (status == B_NO_ERROR) {
		status = pMsg->AddString(kClassNameLabel, kClassName);
	}

	return status;
}

CsoundCVAnalysisJob *
CsoundCVAnalysisJob::Instantiate(BMessage *pMsg)
{
	return validate_instantiation(pMsg, kClassName)
		? new CsoundCVAnalysisJob(pMsg)
		: (CsoundCVAnalysisJob *)0;
}

//	CsoundAbstractJob requirements.
bool
CsoundCVAnalysisJob::IsValid(void) const
{
	// Fourier Analysis requires input & output files.
	return *Filename(kInputFile) != '\0'
		 && *Filename(kOutputFile) != '\0';
}

BPath
CsoundCVAnalysisJob::CurrentDirectory(void) const
{
	// The input file determines the current directory.
	BPath path(Filename(kInputFile));
	if (path.InitCheck() == B_NO_ERROR) {
		path.GetParent(&path);
	}
	return path;
}

char **
CsoundCVAnalysisJob::CopyCommandArguments(
	const char *pFirstArg, int32 &nArgs) const
{
	// Construct argument string from current settings.
	double value;
	char *pCommandLine = CopyString(pFirstArg);
	size_t length = strlen(pCommandLine);
	nArgs = 1;

	length = AppendString(pCommandLine, "-U", 0, length);
	length = AppendString(pCommandLine, "cvanal", 0, length);
	nArgs += 2;
	
	if (!isNaN(value = Number(kSampleRate))) {
		length = AppendString(pCommandLine, "-s", 0, length);
		length = AppendNumber(pCommandLine, "%.0f", value, 0, length);
		nArgs += 2;
	}
	
	if (!isNaN(value = Number(kChannel))) {
		length = AppendString(pCommandLine, "-c", 0, length);
		length = AppendNumber(pCommandLine, "%.0f", value, 0, length);
		nArgs += 2;
	}
	
	if (!isNaN(value = Number(kBeginTime))) {
		length = AppendString(pCommandLine, "-b", 0, length);
		length = AppendNumber(pCommandLine, "%.3f", value, 0, length);
		nArgs += 2;
	}
	
	if (!isNaN(value = Number(kDuration))) {
		length = AppendString(pCommandLine, "-d", 0, length);
		length = AppendNumber(pCommandLine, "%.3f", value, 0, length);
		nArgs += 2;
	}
	
	length = AppendString(pCommandLine, Filename(kInputFile), 0, length);
	length = AppendString(pCommandLine, Filename(kOutputFile), 0, length);
	nArgs += 2;

	// Construct argument vector.
	char **const pArgv = new char *[nArgs + 1];		// accomodate terminating 0
	for (uint32 i = 0; i < nArgs; ++i) {
		pArgv[i] = pCommandLine;
		pCommandLine += strlen(pCommandLine) + 1;
	}
	pArgv[nArgs] = 0;
	
	return pArgv;
}

const char *
CsoundCVAnalysisJob::JobClass(void) const
{
	return kJobClass;
}

char *
CsoundCVAnalysisJob::CopyWindowTitle(void) const
{
	const char *pFileName = JobFilename();
	char *const pTitle =
		new char[strlen(pFileName) + sizeof(kWindowTitleFormat) + 1];
	sprintf(pTitle, kWindowTitleFormat, pFileName);
	return pTitle;
}

CsoundAbstractWindow *
CsoundCVAnalysisJob::CreateWindow(BPoint origin) const
{
	// Factory method: Create an appropriate window.
	char *const pTitle = CopyWindowTitle();
	CsoundAbstractWindow *pWindow =
		new CsoundCVAnalysisMainWindow(WindowPosition(origin), pTitle, *this);
	delete [] pTitle;
	return pWindow;
}
