//		Copyright © 1997–2000  Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//		Abstract class describing a "job"--a collection of settings which
//		describe parameters for running Csound.
//		Each derived job class will have an associated window class, for
//		which it also acts as a factory.

#include <cstdio>
#include <cstring>

#include <Archivable.h>
#include <Entry.h>
#include <File.h>
#include <Message.h>
#include <Node.h>
#include <NodeInfo.h>
#include <Statable.h>
#include <SupportDefs.h>

#include "CsApplication.h"
#include "CsAbstractJob.h"
#include "NaN.h"

int32 CsoundAbstractJob::smJobCount = 0;

// Message fields for archiving.
const char CsoundAbstractJob::kClassNameLabel[] = "class";
const char CsoundAbstractJob::kFilenameLabel[] = "CsFilename";
const char CsoundAbstractJob::kNumberLabel[] = "CsNumber";
const char CsoundAbstractJob::kOptionsLabel[] = "CsOptions";

// Attribute names.
const char kWindowPositionAttr[] = "CsWindowPosition";

// Basic C++ object requirements: construct, copy, assign, destroy.
CsoundAbstractJob::CsoundAbstractJob(
	bool assignName, uint32 numFilenames, uint32 numNumbers, int32 defaultOptions)
:	mRef(),
	mNumFilenames(numFilenames),
	mpFilenames(new char *[numFilenames]),
	mNumNumbers(numNumbers),
	mpNumbers(new double[numNumbers]),
	mOptions(defaultOptions)
{
	if (assignName) {
		// Set up a name for the new job.
		char buffer[100];
		const int32 n = atomic_add(&smJobCount, 1);
		sprintf(buffer, "Untitled %ld", n+1);
		mRef.set_name(buffer);
	}
	
	for (int i = 0; i < mNumFilenames; ++i) {
		mpFilenames[i] = CopyString("");
	}
	
	for (int i = 0; i < mNumNumbers; ++i) {
		mpNumbers[i] = aNaN();
	}
}	

CsoundAbstractJob::CsoundAbstractJob(const CsoundAbstractJob &job)
:	inherited(job),
	mRef(job.mRef),
	mNumFilenames(job.mNumFilenames),
	mpFilenames(new char *[mNumFilenames]),
	mNumNumbers(job.mNumNumbers),
	mpNumbers(new double[mNumNumbers]),
	mOptions(job.mOptions)
{
	for (int i = 0; i < mNumFilenames; ++i) {
		mpFilenames[i] = CopyString(job.mpFilenames[i]);
	}
	
	for (int i = 0; i < mNumNumbers; ++i) {
		mpNumbers[i] = job.mpNumbers[i];
	}
}

CsoundAbstractJob &
CsoundAbstractJob::operator =(const CsoundAbstractJob &job)
{
	if (&job != this) {
		inherited::operator =(job);
		mRef = job.mRef;
		// Assume that mNumFilenames == job.mNumFilenames.
		for (int i = 0; i < mNumFilenames; ++i) {
			delete [] mpFilenames[i];
			mpFilenames[i] = CopyString(job.mpFilenames[i]);
		}
		// Assume that mNumNumbers == job.mNumNumbers.
		for (int i = 0; i < mNumNumbers; ++i) {
			mpNumbers[i] = job.mpNumbers[i];
		}
		mOptions = job.mOptions;
	}
	return *this;
}

CsoundAbstractJob::~CsoundAbstractJob(void)
{
	for (int i = 0; i < mNumFilenames; ++i) {
		delete [] mpFilenames[i];
	}
	delete [] mpFilenames;
	delete [] mpNumbers;
}

bool
CsoundAbstractJob::operator ==(const CsoundAbstractJob &job) const
{
	if (&job == this) {
		return true;
	}
	
	// Don't compare entry_refs.
	
	// Assume that mNumFilenames == job.mNumFilenames.
	for (int i = 0; i < mNumFilenames; ++i) {
		if (0 != strcmp(mpFilenames[i], job.mpFilenames[i])) {
			return false;
		}
	}

	// Assume that mNumNumbers == job.mNumNumbers.
	for (int i = 0; i < mNumNumbers; ++i) {
		const double value1 = mpNumbers[i];
		const double value2 = job.mpNumbers[i];
		if (value1 != value2 && (!isNaN(value1) || !isNaN(value2))) {
			return false;
		}
	}

	return mOptions == job.mOptions;
}

//	BArchivable requirements.  Note that this is an abstract class.
CsoundAbstractJob::CsoundAbstractJob(BMessage *pMsg)
:	inherited(pMsg),
	mRef()
{
	type_code code;
	int32 numEntries;

	if (pMsg->GetInfo(kFilenameLabel, &code, &numEntries) == B_NO_ERROR) {
		mNumFilenames = numEntries;
		mpFilenames = new char *[numEntries];
		const char *pString;
		for (int i = 0; i < numEntries; ++i) {
			if (pMsg->FindString(kFilenameLabel, i, &pString) == B_NO_ERROR) {
				mpFilenames[i] = CopyString(pString);
			} else {
				mpFilenames[i] = CopyString("");
			}
		}
	} else {
		mNumFilenames = 0;
		mpFilenames = 0;
	}
	
	if (pMsg->GetInfo(kNumberLabel, &code, &numEntries) == B_NO_ERROR) {
		mNumNumbers = numEntries;
		mpNumbers = new double[numEntries];
		double value;
		for (int i = 0; i < numEntries; ++i) {
			if (pMsg->FindDouble(kNumberLabel, i, &value) == B_NO_ERROR) {
				mpNumbers[i] = value;
			} else {
				mpNumbers[i] = aNaN();
			}
		}
	} else {
		mNumNumbers = 0;
		mpNumbers = 0;
	}
	
	if (pMsg->FindInt32(kOptionsLabel, &mOptions) != B_NO_ERROR) {
		mOptions = 0;
	}
}

status_t
CsoundAbstractJob::Archive(BMessage *pMsg, bool deep) const
{
	status_t status = inherited::Archive(pMsg, deep);
	int i;
	
	for (i = 0; i < mNumFilenames; ++i) {
		if (status == B_NO_ERROR) {
			status = pMsg->AddString(kFilenameLabel, mpFilenames[i]);
		} else {
			break;
		}
	}

	for (i = 0; i < mNumNumbers; ++i) {
		if (status == B_NO_ERROR) {
			status = pMsg->AddDouble(kNumberLabel, mpNumbers[i]);
		} else {
			break;
		}
	}

	if (status == B_NO_ERROR) {
		status = pMsg->AddInt32(kOptionsLabel, mOptions);
	}

	return status;
}

CsoundAbstractJob *
CsoundAbstractJob::Instantiate(BMessage * /*pMsg*/)
{
	return (CsoundAbstractJob *)0;	// abstract class
}

// Saving and loading.
void
CsoundAbstractJob::SetEntry(const entry_ref &ref)
{
	mRef = ref;
}

const entry_ref *
CsoundAbstractJob::GetEntry(void) const
{
	return &mRef;
}

status_t
CsoundAbstractJob::Save(const entry_ref *pRef)
{
	// If an entry_ref is supplied, first copy it.
	if (pRef) {
		mRef = *pRef;
	}

	// Archive the receiver.
	BMessage msg;
	status_t status = Archive(&msg);
	
	if (status == B_NO_ERROR) {
		// Try to create the file.
		BFile theFile(&mRef, B_WRITE_ONLY|B_CREATE_FILE|B_ERASE_FILE);
		status = theFile.InitCheck();
		if (status == B_NO_ERROR) {
			BNodeInfo nodeInfo(&theFile);
			(void)nodeInfo.SetType(CsoundApplication::GetSignature());
			
			// Flatten the archived message into the file.
			status = msg.Flatten(&theFile);
		}
	}
	
	return status;
}

bool
CsoundAbstractJob::JobFileValid(void) const
{
	BEntry entry(&mRef);
	return entry.Exists() && entry.IsFile();
}

// Filenames.
void
CsoundAbstractJob::SetFilename(int32 which, const char *pName)
{
	if (which >= 0 && which < mNumFilenames) {
		delete [] mpFilenames[which];
		mpFilenames[which] = CopyString(pName ? pName : "");
	}
}

// Numeric parameters.
void
CsoundAbstractJob::SetNumber(int32 which, double value)
{
	if (which >= 0 && which < mNumNumbers) {
		mpNumbers[which] = value;
	}
}

// Preserve the position of the job's window (if it has been saved).
void
CsoundAbstractJob::NoteWindowPosition(BPoint origin) const
{
	if (JobFileValid()) {
		BNode node(&mRef);
		(void)node.WriteAttr(kWindowPositionAttr, B_POINT_TYPE, 0,
									&origin, sizeof(origin));
	}
}

// Methods needed only to load jobs from an earlier version.
void
CsoundAbstractJob::AddFilename(const char *pName)
{
	char **pFilenames = new char *[mNumFilenames + 1];

	for (int i = 0; i < mNumFilenames; ++i) {
		pFilenames[i] = mpFilenames[i];
	}
	pFilenames[mNumFilenames] = CopyString(pName);

	delete [] mpFilenames;
	mpFilenames = pFilenames;
	++mNumFilenames;
}

void
CsoundAbstractJob::AddNumber(double value)
{
	double *pNumbers = new double[mNumNumbers + 1];

	for (int i = 0; i < mNumNumbers; ++i) {
		pNumbers[i] = mpNumbers[i];
	}
	pNumbers[mNumNumbers] = value;

	delete [] mpNumbers;
	mpNumbers = pNumbers;
	++mNumNumbers;
}

// Determine the position of a window about to be created.
BPoint
CsoundAbstractJob::WindowPosition(BPoint suggestedPosition) const
{
	if (JobFileValid()) {
		BNode node(&mRef);
		BPoint saved;
		if (node.ReadAttr(kWindowPositionAttr, B_POINT_TYPE, 0,
								&saved, sizeof(saved)) == sizeof(saved)) {
			return saved;
		}
	}
	
	return suggestedPosition;
}

// String handling.		
char *
CsoundAbstractJob::CopyString(const char *const pChars)
{
	return strcpy(new char[strlen(pChars) + 1], pChars);
}

size_t
CsoundAbstractJob::AppendString(
	char *&pChars1, const char *const pChars2, char separator, size_t oldLength)
{
	if (!pChars1) {
		pChars1 = CopyString(pChars2);
		return strlen(pChars1);
	}

	const size_t len1 = (oldLength ? oldLength : strlen(pChars1));
	const size_t len2 = strlen(pChars2);
	const size_t newLength = len1 + 1 + len2;
	
	char *const pResult = (char *)memcpy(new char[newLength + 1], pChars1, len1);
	pResult[len1] = separator;
	strcpy(pResult+1+len1, pChars2);
	delete [] pChars1;
	pChars1 = pResult;
	
	return newLength;
}

size_t
CsoundAbstractJob::AppendNumber(
	char *&pChars1, const char *const pFormat, double value,
	char separator, size_t oldLength)
{
	char buffer[100];
	sprintf(buffer, pFormat, value);
	return AppendString(pChars1, buffer, separator, oldLength);
}
