//		Copyright © 1997–1999  Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//		Abstract class representing a 'main' window--one which has an associated
//		CsoundAbstractJob, a menu bar, and which is responsible for processing
//		control messages.  A main window may have an additional window for
//		non-essential options.

#ifndef CSABSTRACTMAINWINDOW_H
#define CSABSTRACTMAINWINDOW_H

#include <CheckBox.h>
#include <Entry.h>
#include <FilePanel.h>
#include <MenuItem.h>
#include <Message.h>
#include <Point.h>
#include <PopUpMenu.h>
#include <View.h>

#include "CsAbstractJob.h"
#include "CsAbstractWindow.h"
class CsoundAbstractOptionsWindow;

class CsoundAbstractMainWindow : public CsoundAbstractWindow
{
	typedef CsoundAbstractWindow inherited;
public:

	// Common message codes.
	enum {
		kNewJob							= 'jnew',	// File menu items
		kSaveJob							= 'jsav',
		kSaveAsJob						= 'jsva',
		kRevertJob						= 'jrvt',
		
		kFilenameModified				= 'mfnm',	// filename is being modified
		kSetFilename					= 'sfnm',	// file name changed
		kNumberModified				= 'mnum',	// numeric parameter being modified
		kSetNumber						= 'snum',	// numeric parameter changed
		kSetOption						= 'sopt',	// toggle changed state
		kSetOneOfN						= 's1on',	// 1-of-N option selected

		kShowWindow						= 'show',	// show window
		kShowOptions					= 'shop',	// show options window
		kOptionsWindowClosed			= 'owcl'
	};

	CsoundAbstractMainWindow(BPoint origin, const char *pTitle,
									 const CsoundAbstractJob &job);
	virtual ~CsoundAbstractMainWindow(void);

	// Hook functions.
	virtual void FrameMoved(BPoint origin);
	virtual void MenusBeginning(void);
	virtual void MessageReceived(BMessage *pMsg);
	virtual bool QuitRequested(void);

	// Create BMessages for controls.
	BMessage *CreateMessage(
		const CsoundAbstractWindow::ButtonDescriptor &desc) const;
	BMessage *CreateMessage(
		const CsoundAbstractWindow::FilenameDescriptor &desc,
		bool modifyMessage = false) const;
	BMessage *CreateMessage(
		const CsoundAbstractWindow::NumberDescriptor &desc,
		bool modifyMessage = false) const;
	BMessage *CreateMessage(
		const CsoundAbstractWindow::CheckBoxDescriptor &desc) const;
	BMessage *CreateMessage(
		const CsoundAbstractWindow::RadioButtonDescriptor &desc) const;
	
	// Return or modify current job.
	const CsoundAbstractJob &CurrentJob(void) const;
	CsoundAbstractJob &CurrentJob(void);
	void SetCurrentJob(const CsoundAbstractJob &job);

protected:
	// Return the menu bar.
	BMenuBar *MenuBar(void) const;
	
	// Create a title for an options window.  Caller must delete the string.
	char *CopyOptionsWindowTitle(void);
	
	// Add an options window.  The optional BView parameter specifies
	// vertical alignment for an automatically generated check box.
	void AddOptionsWindow(CsoundAbstractOptionsWindow *pOptionsWindow,
								  float borderSpacing,
								  BView *pRefView = (BView *)0);
	
	// Allow derived classes to react when the job has been modified.
	virtual void UpdateControlValues(void);
	virtual void ValidJob(bool valid) = 0;
	bool IsJobValid(void) const;

	// Accept a dropped file.
	virtual void AcceptDroppedRef(BPoint where, const entry_ref *pRef);

private:
	// Unimplemented copy constructor and assignment operator.
	CsoundAbstractMainWindow(const CsoundAbstractMainWindow &);
	CsoundAbstractMainWindow &operator =(const CsoundAbstractMainWindow &);

	// Process specific messages.
	void ProcessNewJob(BMessage *pMsg);
	void ProcessSaveJob(BMessage *pMsg);
	void ProcessSaveAsJob(BMessage *pMsg);
	void ProcessRevertJob(BMessage *pMsg);
	void ProcessFilePanelClosed(void);
	void ProcessFilenameModified(BMessage *pMsg);
	void ProcessSetFilename(BMessage *pMsg);
	void ProcessNumberModified(BMessage *pMsg);
	void ProcessSetNumber(BMessage *pMsg);
	void ProcessSetOption(BMessage *pMsg);
	void ProcessSetOneOfN(BMessage *pMsg);
	void ProcessDroppedData(BMessage *pMsg);

	// Inform the application about our birth resp. demise.
	void NotifyApplication(uint32 what);
	
	// Find a BTextControl at the given point.
	BTextControl *FindTextControl(BPoint where);

	// Create the menu bar and the common menus.
	BMenuBar *CreateMenuBar(BRect frame);

	// Try to save the current job.
	bool SaveCurrentJob(const BEntry *pEntry = 0);
	
	// Show our file panel.
	void ShowFilePanel(void);

	// Data members.
	CsoundAbstractJob *mpCurrentJob;						// current settings
	CsoundAbstractJob *mpSavedJob;						// last saved settings

	bool mBeingModified;										// some parameter edited
	
	BMenuBar *mpMenuBar;										// the menu bar
	// ### BPopUpMenu *mpSecondaryButtonMenu;			// menus and menu items
	BMenuItem *mpFileSaveItem;
	BMenuItem *mpFileRevertItem;
	BMenu *mpWindowMenu;

	BFilePanel *mpFilePanel;								// file panel for saving
	bool mFileSelectionPending;							// file panel currently open
	
	CsoundAbstractOptionsWindow *mpOptionsWindow;	// secondary controls
	BCheckBox *mpShowOptions;								// show/hide options window
};

inline const CsoundAbstractJob &
CsoundAbstractMainWindow::CurrentJob(void) const
{
	return *mpCurrentJob;
}

inline CsoundAbstractJob &
CsoundAbstractMainWindow::CurrentJob(void)
{
	return *mpCurrentJob;
}

inline BMenuBar *
CsoundAbstractMainWindow::MenuBar(void) const
{
	return mpMenuBar;
}

inline bool
CsoundAbstractMainWindow::IsJobValid(void) const
{
	return !mBeingModified && mpCurrentJob->IsValid();
}

#endif // CSABSTRACTMAINWINDOW_H
