//		Copyright © 1997–2000  Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//		Abstract window class for Csound.  Serves as a base class for all 
//		the dialog-oriented windows.

#include "CsAbstractWindow.h"

#include <cstdio>
#include <cstdlib>
#include <cstring>

#include <Button.h>
#include <CheckBox.h>
#include <Message.h>
#include <Point.h>
#include <Rect.h>
#include <RadioButton.h>
#include <TextControl.h>
#include <View.h>
#include <Window.h>

#include "ColoredTextControl.h"
#include "CsAbstractJob.h"
#include "CsAbstractMainWindow.h"
#include "MenuTextControl.h"
#include "NaN.h"

const rgb_color kMandatoryFilenameColor = { 200, 200, 255, 255 };

CsoundAbstractWindow::CsoundAbstractWindow(
	BPoint origin, const char *pTitle, window_look look, window_feel feel)
:	inherited(BRect(origin, origin), pTitle,
				 look, feel, B_NOT_RESIZABLE|B_NOT_ZOOMABLE),
	mpBackgroundView((BView *)0)
{
	// empty
}

CsoundAbstractWindow::~CsoundAbstractWindow(void)
{
	// empty
}

// Create the window's background view.
void
CsoundAbstractWindow::CreateBackgroundView(
	BRect frame, const char *pTitle, rgb_color viewColor)
{
	mpBackgroundView = new BView(frame, pTitle, B_FOLLOW_NONE, B_WILL_DRAW);
	mpBackgroundView->SetViewColor(viewColor);
	AddChild(mpBackgroundView);
}

//	Utilities for dialog creation.
void
CsoundAbstractWindow::CreateControls(
	const uint32 nItems,
	const CsoundAbstractWindow::ButtonDescriptor descriptors[],
	BButton *pControls[],
	ControlDimensions dimensions[],
	const CsoundAbstractMainWindow &target)
{
	const BRect emptyFrame(B_ORIGIN, B_ORIGIN);
	for (uint32 i = 0; i < nItems; ++i) {
		const ButtonDescriptor &desc = descriptors[i];
		pControls[i] = new BButton(emptyFrame, desc.pName, desc.pLabel,
											target.CreateMessage(desc));
		pControls[i]->SetTarget(&target);
		pControls[i]->ResizeToPreferred();
		pControls[i]->GetPreferredSize(&dimensions[i].width,
												 &dimensions[i].height);
	}
}

void
CsoundAbstractWindow::CreateControls(
	const uint32 nItems,
	const CsoundAbstractWindow::CheckBoxDescriptor descriptors[],
	BCheckBox *pControls[],
	ControlDimensions dimensions[],
	const CsoundAbstractMainWindow &target)
{
	const BRect emptyFrame(B_ORIGIN, B_ORIGIN);
	const int32 options = target.CurrentJob().Options();
	for (uint32 i = 0; i < nItems; ++i) {
		const CheckBoxDescriptor &desc = descriptors[i];
		pControls[i] = new BCheckBox(emptyFrame, desc.pName, desc.pLabel,
											  target.CreateMessage(desc));
		pControls[i]->SetTarget(&target);
		pControls[i]->ResizeToPreferred();
		pControls[i]->GetPreferredSize(&dimensions[i].width,
												 &dimensions[i].height);
	}
	SetControlValues(nItems, descriptors, pControls, target);
}

void
CsoundAbstractWindow::SetControlValues(
	const uint32 nItems,
	const CsoundAbstractWindow::CheckBoxDescriptor descriptors[],
	BCheckBox *pControls[],
	const CsoundAbstractMainWindow &target)
{
	const int32 options = target.CurrentJob().Options();
	for (uint32 i = 0; i < nItems; ++i) {
		const CheckBoxDescriptor &desc = descriptors[i];
		if (desc.option > 0) {
			const bool value = (options & desc.option) != 0;
			pControls[i]->SetValue(value ? B_CONTROL_ON : B_CONTROL_OFF);
		}
	}
}

void
CsoundAbstractWindow::CreateControls(
	const uint32 nItems,
	const CsoundAbstractWindow::RadioButtonDescriptor descriptors[],
	BRadioButton *pControls[],
	ControlDimensions dimensions[],
	const CsoundAbstractMainWindow &target)
{
	const BRect emptyFrame(B_ORIGIN, B_ORIGIN);
	const int32 options = target.CurrentJob().Options();
	for (uint32 i = 0; i < nItems; ++i) {
		const RadioButtonDescriptor &desc = descriptors[i];
		pControls[i] = new BRadioButton(emptyFrame, desc.pName, desc.pLabel,
												  target.CreateMessage(desc));
		pControls[i]->SetTarget(&target);
		pControls[i]->ResizeToPreferred();
		pControls[i]->GetPreferredSize(&dimensions[i].width,
												 &dimensions[i].height);
	}
	SetControlValues(nItems, descriptors, pControls, target);
}

void
CsoundAbstractWindow::SetControlValues(
	const uint32 nItems,
	const CsoundAbstractWindow::RadioButtonDescriptor descriptors[],
	BRadioButton *pControls[],
	const CsoundAbstractMainWindow &target)
{
	const int32 options = target.CurrentJob().Options();
	for (uint32 i = 0; i < nItems; ++i) {
		const RadioButtonDescriptor &desc = descriptors[i];
		if ((options & desc.mask) == desc.value) {
			pControls[i]->SetValue(B_CONTROL_ON);
		}
	}
}

void
CsoundAbstractWindow::CreateControls(
	const uint32 nItems,
	const CsoundAbstractWindow::FilenameDescriptor descriptors[],
	BTextControl *pControls[],
	ControlDimensions dimensions[],
	const CsoundAbstractMainWindow &target)
{
	const BRect emptyFrame(B_ORIGIN, B_ORIGIN);
	for (uint32 i = 0; i < nItems; ++i) {
		const FilenameDescriptor &desc = descriptors[i];
		switch (desc.type) {
		
		case kNormal:
			pControls[i] =
				new BTextControl(emptyFrame, desc.pName, desc.pLabel, "",
									  target.CreateMessage(desc));
			break;

		case kMandatory:
			pControls[i] =
				new ColoredTextControl(emptyFrame, desc.pName, desc.pLabel, "",
											  kMandatoryFilenameColor,
											  target.CreateMessage(desc));
			break;

		case kHasMenu:
			pControls[i] =
				new MenuTextControl(emptyFrame, desc.pName, desc.pLabel, "",
										  target.CreateMessage(desc));
			break;
		}
		
		pControls[i]->SetTarget(&target);
		pControls[i]->SetModificationMessage(target.CreateMessage(desc, true));
		pControls[i]->ResizeToPreferred();
		pControls[i]->GetPreferredSize(&dimensions[i].width,
												 &dimensions[i].height);
		dimensions[i].divider = pControls[i]->Divider();
	}
	SetControlValues(nItems, descriptors, pControls, target);
}

void
CsoundAbstractWindow::SetControlValues(
	const uint32 nItems,
	const CsoundAbstractWindow::FilenameDescriptor descriptors[],
	BTextControl *pControls[],
	const CsoundAbstractMainWindow &target)
{
	for (uint32 i = 0; i < nItems; ++i) {
		pControls[i]->SetText(
			target.CurrentJob().Filename(descriptors[i].whichFile));
	}
}

void
CsoundAbstractWindow::CreateControls(
	const uint32 nItems,
	const CsoundAbstractWindow::NumberDescriptor descriptors[],
	BTextControl *pControls[],
	ControlDimensions dimensions[],
	const CsoundAbstractMainWindow &target)
{
	const BRect emptyFrame(B_ORIGIN, B_ORIGIN);
	for (uint32 i = 0; i < nItems; ++i) {
		const NumberDescriptor &desc = descriptors[i];
		pControls[i] = new BTextControl(emptyFrame, desc.pName, desc.pLabel, "",
												  target.CreateMessage(desc));
		pControls[i]->SetTarget(&target);
		pControls[i]->SetModificationMessage(target.CreateMessage(desc, true));
		pControls[i]->ResizeToPreferred();
		pControls[i]->GetPreferredSize(&dimensions[i].width,
												 &dimensions[i].height);
		dimensions[i].divider = pControls[i]->Divider();
	}
	SetControlValues(nItems, descriptors, pControls, target);
}

void
CsoundAbstractWindow::SetControlValues(
	const uint32 nItems,
	const CsoundAbstractWindow::NumberDescriptor descriptors[],
	BTextControl *pControls[],
	const CsoundAbstractMainWindow &target)
{
	for (uint32 i = 0; i < nItems; ++i) {
		const NumberDescriptor &desc = descriptors[i];
		const double value = target.CurrentJob().Number(desc.whichNumber);
		pControls[i]->SetText(NumberToString(desc.pFormat, value));
	}
}

const char *
CsoundAbstractWindow::NumberToString(const char *pFormat, double value) const
{
	if (isNaN(value)) {
		return "";
	}
	static char buffer[100];
	sprintf(buffer, pFormat, value);
	return buffer;
}

double
CsoundAbstractWindow::StringToNumber(
	const char *pFormat, const char *pString) const
{
	if (pString[0] == '\0') {
		return aNaN();
	}
	// Arrgh.
	return atof(NumberToString(pFormat, atof(pString)));
}
