//		Copyright © 1997–2000  Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//		Describes a Csound heterodyne filter analysis job.

#include <cstdio>
#include <cstring>

#include <Message.h>
#include <Path.h>
#include <Point.h>
#include <SupportDefs.h>

#include "CsAbstractJob.h"
#include "CsHFJob.h"
#include "CsHFMainWindow.h"
#include "NaN.h"

// Class name.
static const char kClassName[] = "CsoundHFAnalysisJob";
static const char kJobClass[] = "HF Analysis";

// Window title prefix.
static const char kWindowTitleFormat[] = "Csound HF Analysis: %s";

// Basic C++ object requirements: construct, copy, assign, destroy.
CsoundHFAnalysisJob::CsoundHFAnalysisJob(
	bool assignName, uint32 numFilenames, uint32 numNumbers)
:	inherited(assignName, numFilenames, numNumbers)
{
	// empty
}	

CsoundHFAnalysisJob::CsoundHFAnalysisJob(const CsoundHFAnalysisJob &job)
:	inherited(job)
{
	// empty
}

CsoundHFAnalysisJob &
CsoundHFAnalysisJob::operator =(const CsoundHFAnalysisJob &job)
{
	if (&job != this) {
		inherited::operator =(job);
	}
	return *this;
}

CsoundHFAnalysisJob::~CsoundHFAnalysisJob(void)
{
	// empty
}

CsoundAbstractJob *
CsoundHFAnalysisJob::New(void) const
{
	// Virtual default constructor.
	return new CsoundHFAnalysisJob();
}

CsoundAbstractJob *
CsoundHFAnalysisJob::Clone(void) const
{
	// Virtual copy constructor.
	return new CsoundHFAnalysisJob(*this);
}

CsoundAbstractJob &
CsoundHFAnalysisJob::Become(const CsoundAbstractJob &job)
{
	// Virtual assignment operator.
	return *this = dynamic_cast<const CsoundHFAnalysisJob &>(job);
}

bool
CsoundHFAnalysisJob::operator ==(const CsoundAbstractJob &abstractJob) const
{
	const CsoundHFAnalysisJob *const pJob =
		dynamic_cast<const CsoundHFAnalysisJob *>(&abstractJob);
	if (!pJob) {
		return false;
	}
	if (pJob == this) {
		return true;
	}
	return inherited::operator ==(abstractJob);
}

//	BArchivable requirements.
CsoundHFAnalysisJob::CsoundHFAnalysisJob(BMessage *pMsg)
:	inherited(pMsg)
{
	// empty
}

status_t
CsoundHFAnalysisJob::Archive(BMessage *pMsg, bool deep) const
{
	status_t status = inherited::Archive(pMsg, deep);

	if (status == B_NO_ERROR) {
		status = pMsg->AddString(kClassNameLabel, kClassName);
	}

	return status;
}

CsoundHFAnalysisJob *
CsoundHFAnalysisJob::Instantiate(BMessage *pMsg)
{
	return validate_instantiation(pMsg, kClassName)
		? new CsoundHFAnalysisJob(pMsg)
		: (CsoundHFAnalysisJob *)0;
}

//	CsoundAbstractJob requirements.
bool
CsoundHFAnalysisJob::IsValid(void) const
{
	// HF Analysis requires input & output files.
	return *Filename(kInputFile) != '\0'
		 && *Filename(kOutputFile) != '\0';
}

BPath
CsoundHFAnalysisJob::CurrentDirectory(void) const
{
	// The input file determines the current directory.
	BPath path(Filename(kInputFile));
	if (path.InitCheck() == B_NO_ERROR) {
		path.GetParent(&path);
	}
	return path;
}

char **
CsoundHFAnalysisJob::CopyCommandArguments(
	const char *pFirstArg, int32 &nArgs) const
{
	// Construct argument string from current settings.
	const char *pFile;
	double value;
	char *pCommandLine = CopyString(pFirstArg);
	size_t length = strlen(pCommandLine);
	nArgs = 1;

	length = AppendString(pCommandLine, "-U", 0, length);
	length = AppendString(pCommandLine, "hetro", 0, length);
	nArgs += 2;
	
	if (*(pFile = Filename(kTranscriptFile))) {
		length = AppendString(pCommandLine, "--", 0, length);
		length = AppendString(pCommandLine, pFile, 0, length);
		nArgs += 2;
	}

	if (!isNaN(value = Number(kSampleRate))) {
		length = AppendString(pCommandLine, "-s", 0, length);
		length = AppendNumber(pCommandLine, "%.0f", value, 0, length);
		nArgs += 2;
	}
	
	if (!isNaN(value = Number(kChannel))) {
		length = AppendString(pCommandLine, "-c", 0, length);
		length = AppendNumber(pCommandLine, "%.0f", value, 0, length);
		nArgs += 2;
	}
	
	if (!isNaN(value = Number(kBeginTime))) {
		length = AppendString(pCommandLine, "-b", 0, length);
		length = AppendNumber(pCommandLine, "%.3f", value, 0, length);
		nArgs += 2;
	}
	
	if (!isNaN(value = Number(kDuration))) {
		length = AppendString(pCommandLine, "-d", 0, length);
		length = AppendNumber(pCommandLine, "%.3f", value, 0, length);
		nArgs += 2;
	}
	
	if (!isNaN(value = Number(kBeginFrequency))) {
		length = AppendString(pCommandLine, "-f", 0, length);
		length = AppendNumber(pCommandLine, "%.0f", value, 0, length);
		nArgs += 2;
	}
	
	if (!isNaN(value = Number(kNumPartials))) {
		length = AppendString(pCommandLine, "-h", 0, length);
		length = AppendNumber(pCommandLine, "%.0f", value, 0, length);
		nArgs += 2;
	}
	
	if (!isNaN(value = Number(kMaxAmplitude))) {
		length = AppendString(pCommandLine, "-M", 0, length);
		length = AppendNumber(pCommandLine, "%.0f", value, 0, length);
		nArgs += 2;
	}
	
	if (!isNaN(value = Number(kMinAmplitude))) {
		length = AppendString(pCommandLine, "-m", 0, length);
		length = AppendNumber(pCommandLine, "%.0f", value, 0, length);
		nArgs += 2;
	}
	
	if (!isNaN(value = Number(kNumBreakpoints))) {
		length = AppendString(pCommandLine, "-n", 0, length);
		length = AppendNumber(pCommandLine, "%.0f", value, 0, length);
		nArgs += 2;
	}
	
	if (!isNaN(value = Number(kLowPassFilter))) {
		length = AppendString(pCommandLine, "-l", 0, length);
		length = AppendNumber(pCommandLine, "%.0f", value, 0, length);
		nArgs += 2;
	}

	length = AppendString(pCommandLine, Filename(kInputFile), 0, length);
	length = AppendString(pCommandLine, Filename(kOutputFile), 0, length);
	nArgs += 2;

	// Construct argument vector.
	char **const pArgv = new char *[nArgs + 1];		// accomodate terminating 0
	for (uint32 i = 0; i < nArgs; ++i) {
		pArgv[i] = pCommandLine;
		pCommandLine += strlen(pCommandLine) + 1;
	}
	pArgv[nArgs] = 0;
	
	return pArgv;
}

const char *
CsoundHFAnalysisJob::JobClass(void) const
{
	return kJobClass;
}

char *
CsoundHFAnalysisJob::CopyWindowTitle(void) const
{
	const char *pFileName = JobFilename();
	char *const pTitle =
		new char[strlen(pFileName) + sizeof(kWindowTitleFormat) + 1];
	sprintf(pTitle, kWindowTitleFormat, pFileName);
	return pTitle;
}

CsoundAbstractWindow *
CsoundHFAnalysisJob::CreateWindow(BPoint origin) const
{
	// Factory method: Create an appropriate window.
	char *const pTitle = CopyWindowTitle();
	CsoundAbstractWindow *pWindow =
		new CsoundHFAnalysisMainWindow(WindowPosition(origin), pTitle, *this);
	delete [] pTitle;
	return pWindow;
}
