//		Copyright © 1997–2000  Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//		Describes a Csound linear predictive analysis job.

#include <cstdio>
#include <cstring>

#include <Message.h>
#include <Path.h>
#include <Point.h>
#include <SupportDefs.h>

#include "CsAbstractJob.h"
#include "CsLPJob.h"
#include "CsLPMainWindow.h"
#include "NaN.h"

// Class name.
static const char kClassName[] = "CsoundLPAnalysisJob";
static const char kJobClass[] = "LP Analysis";

// Window title prefix.
static const char kWindowTitleFormat[] = "Csound LP Analysis: %s";

// Default options.
const int32 kDefaultOptions = CsoundLPAnalysisJob::kSilent;

// Basic C++ object requirements: construct, copy, assign, destroy.
CsoundLPAnalysisJob::CsoundLPAnalysisJob(
	bool assignName, uint32 numFilenames, uint32 numNumbers)
:	inherited(assignName, numFilenames, numNumbers, kDefaultOptions)
{
	// empty
}	

CsoundLPAnalysisJob::CsoundLPAnalysisJob(const CsoundLPAnalysisJob &job)
:	inherited(job)
{
	// empty
}

CsoundLPAnalysisJob &
CsoundLPAnalysisJob::operator =(const CsoundLPAnalysisJob &job)
{
	if (&job != this) {
		inherited::operator =(job);
	}
	return *this;
}

CsoundLPAnalysisJob::~CsoundLPAnalysisJob(void)
{
	// empty
}

CsoundAbstractJob *
CsoundLPAnalysisJob::New(void) const
{
	// Virtual default constructor.
	return new CsoundLPAnalysisJob();
}

CsoundAbstractJob *
CsoundLPAnalysisJob::Clone(void) const
{
	// Virtual copy constructor.
	return new CsoundLPAnalysisJob(*this);
}

CsoundAbstractJob &
CsoundLPAnalysisJob::Become(const CsoundAbstractJob &job)
{
	// Virtual assignment operator.
	return *this = dynamic_cast<const CsoundLPAnalysisJob &>(job);
}

bool
CsoundLPAnalysisJob::operator ==(const CsoundAbstractJob &abstractJob) const
{
	const CsoundLPAnalysisJob *const pJob =
		dynamic_cast<const CsoundLPAnalysisJob *>(&abstractJob);
	if (!pJob) {
		return false;
	}
	if (pJob == this) {
		return true;
	}
	return inherited::operator ==(abstractJob);
}

//	BArchivable requirements.
CsoundLPAnalysisJob::CsoundLPAnalysisJob(BMessage *pMsg)
:	inherited(pMsg)
{
	// empty
}

status_t
CsoundLPAnalysisJob::Archive(BMessage *pMsg, bool deep) const
{
	status_t status = inherited::Archive(pMsg, deep);

	if (status == B_NO_ERROR) {
		status = pMsg->AddString(kClassNameLabel, kClassName);
	}

	return status;
}

CsoundLPAnalysisJob *
CsoundLPAnalysisJob::Instantiate(BMessage *pMsg)
{
	return validate_instantiation(pMsg, kClassName)
		? new CsoundLPAnalysisJob(pMsg)
		: (CsoundLPAnalysisJob *)0;
}

//	CsoundAbstractJob requirements.
bool
CsoundLPAnalysisJob::IsValid(void) const
{
	// LP Analysis requires input & output files.
	return *Filename(kInputFile) != '\0'
		 && *Filename(kOutputFile) != '\0';
}

BPath
CsoundLPAnalysisJob::CurrentDirectory(void) const
{
	// The input file determines the current directory.
	BPath path(Filename(kInputFile));
	if (path.InitCheck() == B_NO_ERROR) {
		path.GetParent(&path);
	}
	return path;
}

char **
CsoundLPAnalysisJob::CopyCommandArguments(
	const char *pFirstArg, int32 &nArgs) const
{
	// Construct argument string from current settings.
	double value;
	const char *pString;
	char *pCommandLine = CopyString(pFirstArg);
	size_t length = strlen(pCommandLine);
	const int32 options = Options();
	nArgs = 1;

	length = AppendString(pCommandLine, "-U", 0, length);
	length = AppendString(pCommandLine, "lpanal", 0, length);
	nArgs += 2;
	
	if (*(pString = Filename(kTranscriptFile))) {
		length = AppendString(pCommandLine, "--", 0, length);
		length = AppendString(pCommandLine, pString, 0, length);
		nArgs += 2;
	}

	if (!isNaN(value = Number(kSampleRate))) {
		length = AppendString(pCommandLine, "-s", 0, length);
		length = AppendNumber(pCommandLine, "%.0f", value, 0, length);
		nArgs += 2;
	}
	
	if (!isNaN(value = Number(kChannel))) {
		length = AppendString(pCommandLine, "-c", 0, length);
		length = AppendNumber(pCommandLine, "%.0f", value, 0, length);
		nArgs += 2;
	}
	
	if (!isNaN(value = Number(kBeginTime))) {
		length = AppendString(pCommandLine, "-b", 0, length);
		length = AppendNumber(pCommandLine, "%.3f", value, 0, length);
		nArgs += 2;
	}
	
	if (!isNaN(value = Number(kDuration))) {
		length = AppendString(pCommandLine, "-d", 0, length);
		length = AppendNumber(pCommandLine, "%.3f", value, 0, length);
		nArgs += 2;
	}
	
	if (!isNaN(value = Number(kNumPoles))) {
		length = AppendString(pCommandLine, "-p", 0, length);
		length = AppendNumber(pCommandLine, "%.0f", value, 0, length);
		nArgs += 2;
	}
	
	if (!isNaN(value = Number(kHopSize))) {
		length = AppendString(pCommandLine, "-h", 0, length);
		length = AppendNumber(pCommandLine, "%.0f", value, 0, length);
		nArgs += 2;
	}
	
	if (*(pString = Filename(kCommentString))) {
		length = AppendString(pCommandLine, "-C", 0, length);
		length = AppendString(pCommandLine, pString, 0, length);
		nArgs += 2;
	}

	if (!isNaN(value = Number(kMinFrequency))) {
		length = AppendString(pCommandLine, "-P", 0, length);
		length = AppendNumber(pCommandLine, "%.0f", value, 0, length);
		nArgs += 2;
	}
	
	if (!isNaN(value = Number(kMaxFrequency))) {
		length = AppendString(pCommandLine, "-Q", 0, length);
		length = AppendNumber(pCommandLine, "%.0f", value, 0, length);
		nArgs += 2;
	}

	length = AppendString(pCommandLine, "-v", 0, length);
	length = AppendNumber(pCommandLine, "%.0f",
								  (double)(options & kVerbosityMask), 0, length);
	nArgs += 2;
	
	if (options & kStorePoles) {
		length = AppendString(pCommandLine, "-a", 0, length);
		nArgs += 1;
	}

	if (options & kDisplays) {
		length = AppendString(pCommandLine, "-g", 0, length);
		nArgs += 1;
	}

	length = AppendString(pCommandLine, Filename(kInputFile), 0, length);
	length = AppendString(pCommandLine, Filename(kOutputFile), 0, length);
	nArgs += 2;

	// Construct argument vector.
	char **const pArgv = new char *[nArgs + 1];		// accomodate terminating 0
	for (uint32 i = 0; i < nArgs; ++i) {
		pArgv[i] = pCommandLine;
		pCommandLine += strlen(pCommandLine) + 1;
	}
	pArgv[nArgs] = 0;
	
	return pArgv;
}

const char *
CsoundLPAnalysisJob::JobClass(void) const
{
	return kJobClass;
}

char *
CsoundLPAnalysisJob::CopyWindowTitle(void) const
{
	const char *pFileName = JobFilename();
	char *const pTitle =
		new char[strlen(pFileName) + sizeof(kWindowTitleFormat) + 1];
	sprintf(pTitle, kWindowTitleFormat, pFileName);
	return pTitle;
}

CsoundAbstractWindow *
CsoundLPAnalysisJob::CreateWindow(BPoint origin) const
{
	// Factory method: Create an appropriate window.
	char *const pTitle = CopyWindowTitle();
	CsoundAbstractWindow *pWindow =
		new CsoundLPAnalysisMainWindow(WindowPosition(origin), pTitle, *this);
	delete [] pTitle;
	return pWindow;
}
