//		Copyright © 1997–2000  Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//		The linear predictive analysis window for Csound.

#include <cstring>
#include <cmath>

#include <Alert.h>
#include <Box.h>
#include <Button.h>
#include <Font.h>
#include <Menu.h>
#include <MenuBar.h>
#include <MenuItem.h>
#include <Message.h>
#include <Point.h>
#include <RadioButton.h>
#include <Rect.h>
#include <TextControl.h>
#include <View.h>

#include "CsAbstractJob.h"
#include "CsLPJob.h"
#include "CsAbstractMainWindow.h"
#include "CsLPMainWindow.h"
#include "CsLPOptionsWindow.h"
#include "CsTranscriptWindow.h"

// Coordinates.
const BPoint kOptionsWindowOffset(80, 60);

// UI description data.
const CsoundAbstractWindow::FilenameDescriptor kFilenameDescriptors[] =
{
	{	"sinp", "Sound Input File", CsoundAbstractWindow::kMandatory,
		CsoundLPAnalysisJob::kInputFile
	},
	{	"outp", "Analysis Output File", CsoundAbstractWindow::kMandatory,
		CsoundLPAnalysisJob::kOutputFile
	},
	{	"tscr", "Transcript", CsoundAbstractWindow::kNormal,
		CsoundLPAnalysisJob::kTranscriptFile
	},
	{	"cmnt", "File Comment", CsoundAbstractWindow::kNormal,
		CsoundLPAnalysisJob::kCommentString
	}
};

const CsoundAbstractWindow::NumberDescriptor kNumberDescriptors[] =
{
	{	"srat", "Sample Rate", "%.0f",
		CsoundLPAnalysisJob::kSampleRate
	},
	{	"chnl", "Channel Number", "%.0f",
		CsoundLPAnalysisJob::kChannel
	},
	{	"btim", "Begin Time", "%.3f",
		CsoundLPAnalysisJob::kBeginTime
	},
	{	"durn", "Duration", "%.3f",
		CsoundLPAnalysisJob::kDuration
	},
	{	"npls", "Number of Poles", "%.0f",
		CsoundLPAnalysisJob::kNumPoles
	},
	{	"hpsz", "Hop Size", "%.0f",
		CsoundLPAnalysisJob::kHopSize
	},
	{	"minf", "Minimum Frequency", "%.0f",
		CsoundLPAnalysisJob::kMinFrequency
	},
	{	"maxf", "Maximum Frequency", "%.0f",
		CsoundLPAnalysisJob::kMaxFrequency
	}
};

const CsoundAbstractWindow::ButtonDescriptor kButtonDescriptors[] =
{
	{	"strt", "Analyze",
		CsoundLPAnalysisMainWindow::kStartAnalysis
	},
	{	"abrt", "Abort",	
		CsoundLPAnalysisMainWindow::kAbortAnalysis
	}
};

CsoundLPAnalysisMainWindow::CsoundLPAnalysisMainWindow(
	BPoint origin, const char *pTitle, const CsoundLPAnalysisJob &job)
:	inherited(origin, pTitle, job),
	mpTranscriptWindow((CsoundTranscriptWindow *)0)
{
	// Create all the controls and determine space requirements.
	uint32 i;
	float maxWidth = 0;
	float maxDivider = 0;
	float maxButtonWidth = 0;

	CsoundAbstractWindow::ControlDimensions filenameDimensions[kNumFilenames];
	CreateControls(kNumFilenames,
						kFilenameDescriptors,
						mpFilenames,
						filenameDimensions,
						*this);
	for (i = 0; i < kNumFilenames; ++i) {
		const CsoundAbstractWindow::ControlDimensions &dim =
			filenameDimensions[i];
		if (dim.width > maxWidth) {
			maxWidth = dim.width;
		}
		if (dim.divider > maxDivider) {
			maxDivider = dim.divider;
		}
	}

	CsoundAbstractWindow::ControlDimensions numberDimensions[kNumNumbers];
	CreateControls(kNumNumbers,
						kNumberDescriptors,
						mpNumbers,
						numberDimensions,
						*this);
	for (i = 0; i < kNumNumbers; ++i) {
		const CsoundAbstractWindow::ControlDimensions &dim =
			numberDimensions[i];
		if (dim.width > maxWidth) {
			maxWidth = dim.width;
		}
		if (dim.divider > maxDivider) {
			maxDivider = dim.divider;
		}
	}

	CsoundAbstractWindow::ControlDimensions buttonDimensions[kNumButtons];
	CreateControls(kNumButtons,
						kButtonDescriptors,
						mpButtons,
						buttonDimensions,
						*this);
	for (i = 0; i < kNumButtons; ++i) {
		if (buttonDimensions[i].width > maxButtonWidth) {
			maxButtonWidth = buttonDimensions[i].width;
		}
	}

	// Calculate overall space requirements and adjust sizes.
	if (maxDivider + 200 > maxWidth) {
		maxWidth = maxDivider + 200;
	}	
	if (kNumButtons * maxButtonWidth > maxWidth) {
		maxWidth = kNumButtons * maxButtonWidth;
	}
	
	for (i = 0; i < kNumFilenames; ++i) {
		BTextControl *pControl = mpFilenames[i];
		// The following had to be empirically determined :-(
		pControl->ResizeTo(maxWidth, filenameDimensions[i].height);
		pControl->SetDivider(maxDivider);
		pControl->TextView()->ResizeTo(maxWidth-maxDivider-10,
												 pControl->TextView()->Frame().Height());
	}
	for (i = 0; i < kNumNumbers; ++i) {
		BTextControl *pControl = mpNumbers[i];
		// The following had to be empirically determined :-(
		pControl->ResizeTo(maxWidth, numberDimensions[i].height);
		pControl->SetDivider(maxDivider);
		pControl->TextView()->ResizeTo(maxWidth-maxDivider-10,
												 pControl->TextView()->Frame().Height());
	}
	for (i = 0; i < kNumButtons; ++i) {
		mpButtons[i]->ResizeTo(maxButtonWidth, buttonDimensions[i].height);
	}
	
	// Lay out the controls.
	BView *const pBackgroundView = BackgroundView();
	
	font_height fontHeight;
	be_bold_font->GetHeight(&fontHeight);
	const float spacing =
		ceil((fontHeight.ascent + fontHeight.descent + fontHeight.leading) / 2);
	
	float y = 2 * spacing;
	for (i = 0; i <= kTranscriptFilename; ++i) {
		mpFilenames[i]->MoveTo(spacing, y);
		y += filenameDimensions[i].height;
	}
	BRect filesFrame(B_ORIGIN, BPoint(maxWidth + 2 * spacing, y + spacing));
	filesFrame.OffsetBy(BPoint(spacing, spacing));
	BBox *const pFiles = new BBox(filesFrame, "files");
	pFiles->SetLabel("Files");
	for (i = 0; i <= kTranscriptFilename; ++i) {
		pFiles->AddChild(mpFilenames[i]);
	}
	pBackgroundView->AddChild(pFiles);

	y = 2 * spacing;
	mpFilenames[kCommentString]->MoveTo(spacing, y);
	y += filenameDimensions[kCommentString].height;
	for (i = 0; i < kNumNumbers; ++i) {
		mpNumbers[i]->MoveTo(spacing, y);
		y += numberDimensions[i].height;
	}
	BRect paramsFrame(B_ORIGIN, BPoint(maxWidth + 2 * spacing, y + spacing));
	paramsFrame.OffsetBy(pFiles->Frame().LeftBottom());
	paramsFrame.OffsetBy(BPoint(0, spacing));
	BBox *const pParams = new BBox(paramsFrame, "params");
	pParams->SetLabel("Parameters");
	pParams->AddChild(mpFilenames[kCommentString]);
	for (i = 0; i < kNumNumbers; ++i) {
		pParams->AddChild(mpNumbers[i]);
	}
	pBackgroundView->AddChild(pParams);
	
	BButton *pStart = mpButtons[kStart];
	BButton *pAbort = mpButtons[kAbort];
	pStart->MoveTo(spacing, pParams->Frame().bottom + spacing);
	pAbort->MoveTo(pStart->Frame().RightTop() + BPoint(spacing, 0));
	pBackgroundView->AddChild(pStart);
	pBackgroundView->AddChild(pAbort);
	
	const float preferredWidth = maxWidth + 4 * spacing;
	const float preferredHeight = pStart->Frame().bottom + spacing;

	// Resize to final dimensions.
	pBackgroundView->ResizeTo(preferredWidth, preferredHeight);
	ResizeTo(preferredWidth, pBackgroundView->Frame().bottom);
	
	// Initialize control states.
	pStart->MakeDefault(true);
	pStart->SetEnabled(IsJobValid());
	pAbort->SetEnabled(false);

	// Create options window.
	char *const pOptionsTitle = CopyOptionsWindowTitle();
	CsoundLPAnalysisOptionsWindow	 *pOptions =
		new CsoundLPAnalysisOptionsWindow(
				origin + kOptionsWindowOffset, pOptionsTitle, *this);
	delete [] pOptionsTitle;
	AddOptionsWindow(pOptions, spacing, pAbort);
	pOptions->Run();
}

CsoundLPAnalysisMainWindow::~CsoundLPAnalysisMainWindow(void)
{
	// empty
}

void
CsoundLPAnalysisMainWindow::MessageReceived(BMessage *pMsg)
{
	switch (pMsg->what) {

	// Command button messages.
	case kStartAnalysis:
		if (!mpTranscriptWindow && IsJobValid()) {
			// Start the job in a new transcript window.
			mpTranscriptWindow =
				new CsoundTranscriptWindow(Frame().LeftTop() + BPoint(100, 100),
													 Title(), *this, "csound", CurrentJob());
			mpButtons[kStart]->SetEnabled(false);
			mpButtons[kAbort]->SetEnabled(true);			
		}
		break;
		
	case kAbortAnalysis:
		if (mpTranscriptWindow) {
			mpTranscriptWindow->PostMessage(CsoundTranscriptWindow::kAbortJob);
		}
		break;
		
	case CsoundTranscriptWindow::kJobFailed:
		if (mpTranscriptWindow && mpTranscriptWindow->Lock()) {
			mpTranscriptWindow->Quit();
		}
		(new BAlert("", "Failed to start analysis.", "Duh!"))->Go();
		// FALL THROUGH
	case CsoundTranscriptWindow::kJobFinished:
		mpTranscriptWindow = (CsoundTranscriptWindow *)0;	
		mpButtons[kStart]->SetEnabled(IsJobValid());
		mpButtons[kAbort]->SetEnabled(false);
		break;

	// Fall back.
	default:
		inherited::MessageReceived(pMsg);
		break;
	}
}

void
CsoundLPAnalysisMainWindow::UpdateControlValues(void)
{
	// Update all controls to their new values.
	SetControlValues(kNumFilenames, kFilenameDescriptors, mpFilenames, *this);
	SetControlValues(kNumNumbers, kNumberDescriptors, mpNumbers, *this);
	
	inherited::UpdateControlValues();
}

void
CsoundLPAnalysisMainWindow::ValidJob(bool valid)
{
	if (!mpTranscriptWindow) {
		mpButtons[kStart]->SetEnabled(valid);
	}
}
