//		Copyright © 1997–2000  Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//		The linear predictive analysis options window;
//		contains less frequently used settings.

#include <cmath>

#include <Box.h>
#include <CheckBox.h>
#include <Font.h>
#include <Point.h>
#include <RadioButton.h>
#include <Rect.h>

#include "CsLPJob.h"
#include "CsLPMainWindow.h"
#include "CsLPOptionsWindow.h"

// UI description data.
const CsoundAbstractWindow::CheckBoxDescriptor kCheckBoxDescriptors[] =
{
	{	"stpl", "Store Filter Poles",
		CsoundLPAnalysisJob::kStorePoles
	},
	{	"disp", "Graphical Displays",
		CsoundLPAnalysisJob::kDisplays
	}
};

const CsoundAbstractWindow::RadioButtonDescriptor kRadioButtonDescriptors[] =
{
	{	"slnt", "Silent",
		CsoundLPAnalysisJob::kVerbosityMask,
		CsoundLPAnalysisJob::kSilent
	},
	{	"vrbs", "Verbose",
		CsoundLPAnalysisJob::kVerbosityMask,
		CsoundLPAnalysisJob::kVerbose
	},
	{	"dbug", "Debug",
		CsoundLPAnalysisJob::kVerbosityMask,
		CsoundLPAnalysisJob::kDebug
	}
};

CsoundLPAnalysisOptionsWindow::CsoundLPAnalysisOptionsWindow(
	BPoint origin, const char *pTitle, CsoundLPAnalysisMainWindow &target)
:	inherited(origin, pTitle, target)
{
	// Create all the controls and determine space requirements.
	uint32 i;
	float maxWidth = 0;
	float maxCheckBoxWidth = 0;
	float maxRadioButtonWidth = 0;

	CsoundAbstractWindow::ControlDimensions checkBoxDimensions[kNumCheckBoxes];
	CreateControls(kNumCheckBoxes,
						kCheckBoxDescriptors,
						mpCheckBoxes,
						checkBoxDimensions,
						target);
	for (i = 0; i < kNumCheckBoxes; ++i) {
		const CsoundAbstractWindow::ControlDimensions &dim =
			checkBoxDimensions[i];
		if (dim.width > maxWidth) {
			maxWidth = dim.width;
		}
		if (dim.width > maxCheckBoxWidth) {
			maxCheckBoxWidth = dim.width;
		}
	}

	CsoundAbstractWindow::ControlDimensions radioButtonDimensions[kNumRadioButtons];
	CreateControls(kNumRadioButtons,
						kRadioButtonDescriptors,
						mpRadioButtons,
						radioButtonDimensions,
						target);
	for (i = 0; i < kNumRadioButtons; ++i) {
		const CsoundAbstractWindow::ControlDimensions &dim =
			radioButtonDimensions[i];
		if (dim.width > maxWidth) {
			maxWidth = dim.width;
		}
		if (dim.width > maxRadioButtonWidth) {
			maxRadioButtonWidth = dim.width;
		}
	}

	// Calculate overall space requirements and adjust sizes.
	font_height fontHeight;
	be_bold_font->GetHeight(&fontHeight);
	const float spacing =
		ceil((fontHeight.ascent + fontHeight.descent + fontHeight.leading) / 2);
	
	if (maxCheckBoxWidth + maxRadioButtonWidth + 3 * spacing > maxWidth) {
		maxWidth = maxCheckBoxWidth + maxRadioButtonWidth + 3 * spacing;
	}

	// Lay out the controls.
	BView *const pBackgroundView = BackgroundView();
	
	float y = 2 * spacing;
	for (i = 0; i < kNumRadioButtons; ++i) {
		mpRadioButtons[i]->MoveTo(spacing, y);
		y += radioButtonDimensions[i].height;
	}
	BRect messagesFrame(B_ORIGIN,
							  BPoint(maxRadioButtonWidth + 2 * spacing, y + spacing));
	messagesFrame.OffsetBy(BPoint(spacing, spacing));
	BBox *const pMessages = new BBox(messagesFrame, "messages");
	pMessages->SetLabel("Messages");
	for (i = 0; i < kNumRadioButtons; ++i) {
		pMessages->AddChild(mpRadioButtons[i]);
	}
	pBackgroundView->AddChild(pMessages);

	float y1 = 2 * spacing;
	for (i = 0; i < kNumCheckBoxes; ++i) {
		mpCheckBoxes[i]->MoveTo(spacing, y1);
		y1 += checkBoxDimensions[i].height;
	}
	BRect switchesFrame(B_ORIGIN,
							  BPoint(maxCheckBoxWidth + 2 * spacing, y + spacing));
	switchesFrame.OffsetBy(pMessages->Frame().RightTop());
	switchesFrame.OffsetBy(BPoint(spacing, 0));
	BBox *const pSwitches = new BBox(switchesFrame, "switches");
	pSwitches->SetLabel("Switches");
	for (i = 0; i < kNumCheckBoxes; ++i) {
		pSwitches->AddChild(mpCheckBoxes[i]);
	}
	pBackgroundView->AddChild(pSwitches);

	// Resize to final dimensions.
	const float preferredWidth = maxWidth + 4 * spacing;
	const float preferredHeight = pMessages->Frame().bottom + spacing;
	pBackgroundView->ResizeTo(preferredWidth, preferredHeight);
	ResizeTo(preferredWidth, pBackgroundView->Frame().bottom);
}

CsoundLPAnalysisOptionsWindow::~CsoundLPAnalysisOptionsWindow(void)
{
	// empty
}

void
CsoundLPAnalysisOptionsWindow::UpdateControlValues(void)
{
	// Update all controls to their new values.
	SetControlValues(kNumCheckBoxes,
						  kCheckBoxDescriptors,
						  mpCheckBoxes,
						  Target());
	SetControlValues(kNumRadioButtons,
						  kRadioButtonDescriptors,
						  mpRadioButtons,
						  Target());
}
