//		Copyright © 1997–2000  Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//		The fourier (PVOC) analysis window for Csound.

#include <cstring>
#include <cmath>

#include <Alert.h>
#include <Box.h>
#include <Button.h>
#include <CheckBox.h>
#include <Font.h>
#include <Menu.h>
#include <MenuBar.h>
#include <MenuItem.h>
#include <Message.h>
#include <Point.h>
#include <Rect.h>
#include <TextControl.h>
#include <View.h>

#include "CsAbstractJob.h"
#include "CsPVJob.h"
#include "CsAbstractMainWindow.h"
#include "CsPVMainWindow.h"
#include "CsTranscriptWindow.h"

// UI description data.
const CsoundAbstractWindow::FilenameDescriptor kFilenameDescriptors[] =
{
	{	"sinp", "Sound Input File", CsoundAbstractWindow::kMandatory,
		CsoundPVAnalysisJob::kInputFile
	},
	{	"outp", "Analysis Output File", CsoundAbstractWindow::kMandatory,
		CsoundPVAnalysisJob::kOutputFile
	},
	{	"tscr", "Transcript", CsoundAbstractWindow::kNormal,
		CsoundPVAnalysisJob::kTranscriptFile
	},
	{	"trce", "Verbose Output", CsoundAbstractWindow::kNormal,
		CsoundPVAnalysisJob::kTraceOutputFile
	}
};

const CsoundAbstractWindow::NumberDescriptor kNumberDescriptors[] =
{
	{	"srat", "Sample Rate", "%.0f",
		CsoundPVAnalysisJob::kSampleRate
	},
	{	"chnl", "Channel Number", "%.0f",
		CsoundPVAnalysisJob::kChannel
	},
	{	"btim", "Begin Time", "%.3f",
		CsoundPVAnalysisJob::kBeginTime
	},
	{	"durn", "Duration", "%.3f",
		CsoundPVAnalysisJob::kDuration
	},
	{	"frsz", "Frame Size", "%.0f",
		CsoundPVAnalysisJob::kFrameSize
	},
	{	"wolp", "Window Overlap", "%.0f",
		CsoundPVAnalysisJob::kWindowOverlap
	},
	{	"hpsz", "Hop Size", "%.0f",
		CsoundPVAnalysisJob::kHopSize
	}
};

const CsoundAbstractWindow::CheckBoxDescriptor kCheckBoxDescriptors[] =
{
	{	"disp", "Graphical Displays",
		CsoundPVAnalysisJob::kDisplays
	}
};

const CsoundAbstractWindow::ButtonDescriptor kButtonDescriptors[] =
{
	{	"strt", "Analyze",
		CsoundPVAnalysisMainWindow::kStartAnalysis
	},
	{	"abrt", "Abort",	
		CsoundPVAnalysisMainWindow::kAbortAnalysis
	}
};

CsoundPVAnalysisMainWindow::CsoundPVAnalysisMainWindow(
	BPoint origin, const char *pTitle, const CsoundPVAnalysisJob &job)
:	inherited(origin, pTitle, job),
	mpTranscriptWindow((CsoundTranscriptWindow *)0)
{
	// Create all the controls and determine space requirements.
	uint32 i;
	float maxWidth = 0;
	float maxDivider = 0;
	float maxCheckBoxWidth = 0;
	float maxButtonWidth = 0;

	CsoundAbstractWindow::ControlDimensions filenameDimensions[kNumFilenames];
	CreateControls(kNumFilenames,
						kFilenameDescriptors,
						mpFilenames,
						filenameDimensions,
						*this);
	for (i = 0; i < kNumFilenames; ++i) {
		const CsoundAbstractWindow::ControlDimensions &dim =
			filenameDimensions[i];
		if (dim.width > maxWidth) {
			maxWidth = dim.width;
		}
		if (dim.divider > maxDivider) {
			maxDivider = dim.divider;
		}
	}

	CsoundAbstractWindow::ControlDimensions numberDimensions[kNumNumbers];
	CreateControls(kNumNumbers,
						kNumberDescriptors,
						mpNumbers,
						numberDimensions,
						*this);
	for (i = 0; i < kNumNumbers; ++i) {
		const CsoundAbstractWindow::ControlDimensions &dim =
			numberDimensions[i];
		if (dim.width > maxWidth) {
			maxWidth = dim.width;
		}
		if (dim.divider > maxDivider) {
			maxDivider = dim.divider;
		}
	}

	CsoundAbstractWindow::ControlDimensions checkBoxDimensions[kNumCheckBoxes];
	CreateControls(kNumCheckBoxes,
						kCheckBoxDescriptors,
						mpCheckBoxes,
						checkBoxDimensions,
						*this);
	for (i = 0; i < kNumCheckBoxes; ++i) {
		const CsoundAbstractWindow::ControlDimensions &dim =
			checkBoxDimensions[i];
		if (dim.width > maxWidth) {
			maxWidth = dim.width;
		}
		if (dim.width > maxCheckBoxWidth) {
			maxCheckBoxWidth = dim.width;
		}
	}

	CsoundAbstractWindow::ControlDimensions buttonDimensions[kNumButtons];
	CreateControls(kNumButtons,
						kButtonDescriptors,
						mpButtons,
						buttonDimensions,
						*this);
	for (i = 0; i < kNumButtons; ++i) {
		if (buttonDimensions[i].width > maxButtonWidth) {
			maxButtonWidth = buttonDimensions[i].width;
		}
	}

	// Calculate overall space requirements and adjust sizes.
	if (maxDivider + 200 > maxWidth) {
		maxWidth = maxDivider + 200;
	}	
	if (maxCheckBoxWidth > maxWidth) {
		maxWidth = maxCheckBoxWidth;
	}
	if (kNumButtons * maxButtonWidth > maxWidth) {
		maxWidth = kNumButtons * maxButtonWidth;
	}
	
	for (i = 0; i < kNumFilenames; ++i) {
		BTextControl *pControl = mpFilenames[i];
		// The following had to be empirically determined :-(
		pControl->ResizeTo(maxWidth, filenameDimensions[i].height);
		pControl->SetDivider(maxDivider);
		pControl->TextView()->ResizeTo(maxWidth-maxDivider-10,
												 pControl->TextView()->Frame().Height());
	}
	for (i = 0; i < kNumNumbers; ++i) {
		BTextControl *pControl = mpNumbers[i];
		// The following had to be empirically determined :-(
		pControl->ResizeTo(maxWidth, numberDimensions[i].height);
		pControl->SetDivider(maxDivider);
		pControl->TextView()->ResizeTo(maxWidth-maxDivider-10,
												 pControl->TextView()->Frame().Height());
	}
	for (i = 0; i < kNumButtons; ++i) {
		mpButtons[i]->ResizeTo(maxButtonWidth, buttonDimensions[i].height);
	}
	
	// Lay out the controls.
	BView *const pBackgroundView = BackgroundView();
	
	font_height fontHeight;
	be_bold_font->GetHeight(&fontHeight);
	const float spacing =
		ceil((fontHeight.ascent + fontHeight.descent + fontHeight.leading) / 2);
	
	float y = 2 * spacing;
	for (i = 0; i < kNumFilenames; ++i) {
		mpFilenames[i]->MoveTo(spacing, y);
		y += filenameDimensions[i].height;
	}
	BRect filesFrame(B_ORIGIN, BPoint(maxWidth + 2 * spacing, y + spacing));
	filesFrame.OffsetBy(BPoint(spacing, spacing));
	BBox *const pFiles = new BBox(filesFrame, "files");
	pFiles->SetLabel("Files");
	for (i = 0; i < kNumFilenames; ++i) {
		pFiles->AddChild(mpFilenames[i]);
	}
	pBackgroundView->AddChild(pFiles);

	y = 2 * spacing;
	for (i = 0; i < kNumNumbers; ++i) {
		mpNumbers[i]->MoveTo(spacing, y);
		y += numberDimensions[i].height;
	}
	BRect numbersFrame(B_ORIGIN, BPoint(maxWidth + 2 * spacing, y + spacing));
	numbersFrame.OffsetBy(pFiles->Frame().LeftBottom());
	numbersFrame.OffsetBy(BPoint(0, spacing));
	BBox *const pNumbers = new BBox(numbersFrame, "numbers");
	pNumbers->SetLabel("Parameters");
	for (i = 0; i < kNumNumbers; ++i) {
		pNumbers->AddChild(mpNumbers[i]);
	}
	pBackgroundView->AddChild(pNumbers);
	
	y = 2 * spacing;
	for (i = 0; i < kNumCheckBoxes; ++i) {
		mpCheckBoxes[i]->MoveTo(spacing, y);
		y += checkBoxDimensions[i].height;
	}
	BRect switchesFrame(B_ORIGIN, BPoint(maxWidth + 2 * spacing, y + spacing));
	switchesFrame.OffsetBy(pNumbers->Frame().LeftBottom());
	switchesFrame.OffsetBy(BPoint(0, spacing));
	BBox *const pSwitches = new BBox(switchesFrame, "switches");
	pSwitches->SetLabel("Switches");
	for (i = 0; i < kNumCheckBoxes; ++i) {
		pSwitches->AddChild(mpCheckBoxes[i]);
	}
	pBackgroundView->AddChild(pSwitches);

	BButton *pStart = mpButtons[kStart];
	BButton *pAbort = mpButtons[kAbort];
	pStart->MoveTo(spacing, pSwitches->Frame().bottom + spacing);
	pAbort->MoveTo(pStart->Frame().RightTop() + BPoint(spacing, 0));
	pBackgroundView->AddChild(pStart);
	pBackgroundView->AddChild(pAbort);
	
	const float preferredWidth = maxWidth + 4 * spacing;
	const float preferredHeight = pStart->Frame().bottom + spacing;

	// Resize to final dimensions.
	pBackgroundView->ResizeTo(preferredWidth, preferredHeight);
	ResizeTo(preferredWidth, pBackgroundView->Frame().bottom);
	
	// Initialize control states.
	pStart->MakeDefault(true);
	pStart->SetEnabled(IsJobValid());
	pAbort->SetEnabled(false);
}

CsoundPVAnalysisMainWindow::~CsoundPVAnalysisMainWindow(void)
{
	// empty
}

void
CsoundPVAnalysisMainWindow::MessageReceived(BMessage *pMsg)
{
	switch (pMsg->what) {

	// Command button messages.
	case kStartAnalysis:
		if (!mpTranscriptWindow && IsJobValid()) {
			// Start the job in a new transcript window.
			mpTranscriptWindow =
				new CsoundTranscriptWindow(Frame().LeftTop() + BPoint(100, 100),
													 Title(), *this, "csound", CurrentJob());
			mpButtons[kStart]->SetEnabled(false);
			mpButtons[kAbort]->SetEnabled(true);			
		}
		break;
		
	case kAbortAnalysis:
		if (mpTranscriptWindow) {
			mpTranscriptWindow->PostMessage(CsoundTranscriptWindow::kAbortJob);
		}
		break;
		
	case CsoundTranscriptWindow::kJobFailed:
		if (mpTranscriptWindow && mpTranscriptWindow->Lock()) {
			mpTranscriptWindow->Quit();
		}
		(new BAlert("", "Failed to start analysis.", "Duh!"))->Go();
		// FALL THROUGH
	case CsoundTranscriptWindow::kJobFinished:
		mpTranscriptWindow = (CsoundTranscriptWindow *)0;	
		mpButtons[kStart]->SetEnabled(IsJobValid());
		mpButtons[kAbort]->SetEnabled(false);
		break;

	// Fall back.
	default:
		inherited::MessageReceived(pMsg);
		break;
	}
}

void
CsoundPVAnalysisMainWindow::UpdateControlValues(void)
{
	// Update all controls to their new values.
	SetControlValues(kNumFilenames, kFilenameDescriptors, mpFilenames, *this);
	SetControlValues(kNumNumbers, kNumberDescriptors, mpNumbers, *this);
	SetControlValues(kNumCheckBoxes, kCheckBoxDescriptors, mpCheckBoxes, *this);
		
	inherited::UpdateControlValues();
}

void
CsoundPVAnalysisMainWindow::ValidJob(bool valid)
{
	if (!mpTranscriptWindow) {
		mpButtons[kStart]->SetEnabled(valid);
	}
}
