//		Copyright © 1997–2000  Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//		Describes a Csound synthesis job--the normal mode of running Csound.

#include <cstdio>
#include <cstring>

#include <Message.h>
#include <Path.h>
#include <Point.h>
#include <SupportDefs.h>

#include "CsAbstractJob.h"
#include "CsSynthJob.h"
#include "CsSynthMainWindow.h"
#include "NaN.h"

// Class name.
static const char kClassName[] = "CsoundSynthesisJob";
static const char kJobClass[] = "Synthesis";

// Window title prefix.
static const char kWindowTitleFormat[] = "Csound Synthesis: %s";

// Default options.
const int32 kDefaultOptions =
	 CsoundSynthesisJob::kAIFFFormat
	|CsoundSynthesisJob::kSigned16BitSamples
	|CsoundSynthesisJob::kNoteAmplitudeMsgs
	|CsoundSynthesisJob::kOutOfRangeMsgs
	|CsoundSynthesisJob::kWarningMsgs
	|CsoundSynthesisJob::kGraphicsDisplays;

// Basic C++ object requirements: construct, copy, assign, destroy.
CsoundSynthesisJob::CsoundSynthesisJob(bool assignName)
:	inherited(assignName, kNumFilenames, kNumNumbers, kDefaultOptions)
{
	// empty
}	

CsoundSynthesisJob::CsoundSynthesisJob(const CsoundSynthesisJob &job)
:	inherited(job)
{
	// empty
}

CsoundSynthesisJob &
CsoundSynthesisJob::operator =(const CsoundSynthesisJob &job)
{
	if (&job != this) {
		inherited::operator =(job);
	}
	return *this;
}

CsoundSynthesisJob::~CsoundSynthesisJob(void)
{
	// empty
}

CsoundAbstractJob *
CsoundSynthesisJob::New(void) const
{
	// Virtual default constructor.
	return new CsoundSynthesisJob();
}

CsoundAbstractJob *
CsoundSynthesisJob::Clone(void) const
{
	// Virtual copy constructor.
	return new CsoundSynthesisJob(*this);
}

CsoundAbstractJob &
CsoundSynthesisJob::Become(const CsoundAbstractJob &job)
{
	// Virtual assignment operator.
	return *this = dynamic_cast<const CsoundSynthesisJob &>(job);
}

bool
CsoundSynthesisJob::operator ==(const CsoundAbstractJob &abstractJob) const
{
	const CsoundSynthesisJob *const pJob =
		dynamic_cast<const CsoundSynthesisJob *>(&abstractJob);
	if (!pJob) {
		return false;
	}
	if (pJob == this) {
		return true;
	}
	return inherited::operator ==(abstractJob);
}

//	BArchivable requirements.
CsoundSynthesisJob::CsoundSynthesisJob(BMessage *pMsg)
:	inherited(pMsg)
{
	// Update from earlier versions.
	if (NumFilenames() == kMIDIOutFile) {
		AddFilename("");
	}
}

status_t
CsoundSynthesisJob::Archive(BMessage *pMsg, bool deep) const
{
	status_t status = inherited::Archive(pMsg, deep);

	if (status == B_NO_ERROR) {
		status = pMsg->AddString(kClassNameLabel, kClassName);
	}

	return status;
}

CsoundSynthesisJob *
CsoundSynthesisJob::Instantiate(BMessage *pMsg)
{
	return validate_instantiation(pMsg, kClassName)
		? new CsoundSynthesisJob(pMsg)
		: (CsoundSynthesisJob *)0;
}

//	CsoundAbstractJob requirements.
bool
CsoundSynthesisJob::IsValid(void) const
{
	// Synthesis requires orchestra & score files.
	return *Filename(kOrchestraFile) != '\0'
		 && *Filename(kScoreFile) != '\0';
}

BPath
CsoundSynthesisJob::CurrentDirectory(void) const
{
	// The orchestra file determines the current directory.
	BPath path(Filename(kOrchestraFile));
	if (path.InitCheck() == B_NO_ERROR) {
		path.GetParent(&path);
	}
	return path;
}

char **
CsoundSynthesisJob::CopyCommandArguments(
	const char *pFirstArg, int32 &nArgs) const
{
	// Construct argument string from current settings.
	const char *pFile;
	double value;
	char *pCommandLine = CopyString(pFirstArg);
	size_t length = strlen(pCommandLine);
	const int32 options = Options();
	nArgs = 1;
	
	if (options & kPreprocessScore) {
		length = AppendString(pCommandLine, "-C", 0, length);
		nArgs += 1;
	}

//	if (options & kTranslateScore) {
//		length = AppendString(pCommandLine, "-S", 0, length);
//		nArgs += 1;
//	}

	if (*(pFile = Filename(kExtractFile))) {
		length = AppendString(pCommandLine, "-x", 0, length);
		length = AppendString(pCommandLine, pFile, 0, length);
		nArgs += 2;
	}

	if (*(pFile = Filename(kInputFile))) {
		length = AppendString(pCommandLine, "-i", 0, length);
		length = AppendString(pCommandLine, pFile, 0, length);
		nArgs += 2;
	}

	if (*(pFile = Filename(kLineFile))) {
		length = AppendString(pCommandLine, "-L", 0, length);
		length = AppendString(pCommandLine, pFile, 0, length);
		nArgs += 2;
	}

	if (*(pFile = Filename(kMIDIInFile))) {
		length = AppendString(pCommandLine,
									 (options & kRealtimeMIDI) ? "-M" : "-F",
									 0, length);
		length = AppendString(pCommandLine, pFile, 0, length);
		nArgs += 2;
	}

	if (*(pFile = Filename(kMIDIOutFile))) {
		length = AppendString(pCommandLine, "-Q", 0, length);
		length = AppendString(pCommandLine, pFile, 0, length);
		nArgs += 2;
	}

	if (*(pFile = Filename(kOutputFile))) {
		length = AppendString(pCommandLine, "-o", 0, length);
		length = AppendString(pCommandLine, pFile, 0, length);
		nArgs += 2;
	}

	if (*(pFile = Filename(kTranscriptFile))) {
		length = AppendString(pCommandLine, "--", 0, length);
		length = AppendString(pCommandLine, pFile, 0, length);
		nArgs += 2;
	}

	if (!isNaN(value = Number(kSoftwareSampleFrames))) {
		//length = AppendString(pCommandLine, "-b", 0, length);
		//length = AppendNumber(pCommandLine, "%.0f", value, 0, length);
		//nArgs += 2;
		length = AppendNumber(pCommandLine, "-b%.0f", value, 0, length);
		nArgs += 1;
	}

	if (!isNaN(value = Number(kHardwareSampleFrames))) {
		length = AppendString(pCommandLine, "-B", 0, length);
		length = AppendNumber(pCommandLine, "%.0f", value, 0, length);
		nArgs += 2;
	}

	if (!isNaN(value = Number(kSampleRateOverride))) {
		length = AppendString(pCommandLine, "-r", 0, length);
		length = AppendNumber(pCommandLine, "%.0f", value, 0, length);
		nArgs += 2;
	}

	if (!isNaN(value = Number(kControlRateOverride))) {
		length = AppendString(pCommandLine, "-k", 0, length);
		length = AppendNumber(pCommandLine, "%.0f", value, 0, length);
		nArgs += 2;
	}

	if (!isNaN(value = Number(kTempo))) {
		length = AppendString(pCommandLine, "-t", 0, length);
		length = AppendNumber(pCommandLine, "%.0f", value, 0, length);
		nArgs += 2;
	}

//	if (!isNaN(value = Number(kSustainPedalThreshold))) {
//		length = AppendString(pCommandLine, "-P", 0, length);
//		length = AppendNumber(pCommandLine, "%.0f", value, 0, length);
//		nArgs += 2;
//	}
	
	if (options & kInitTimeOnly) {
		length = AppendString(pCommandLine, "-I", 0, length);
		nArgs += 1;
	}

	if (options & kInhibitOutput) {
		length = AppendString(pCommandLine, "-n", 0, length);
		nArgs += 1;
	}

	switch (options & kFileFormatMask) {

	case kNoHeaderFormat:
		length = AppendString(pCommandLine, "-h", 0, length);
		nArgs += 1;
		break;
	
	case kSoundfileFormat:
		length = AppendString(pCommandLine, "-J", 0, length);
		nArgs += 1;
		break;
	
	case kAIFFFormat:
		length = AppendString(pCommandLine, "-A", 0, length);
		nArgs += 1;
		break;
	
	case kWAVFormat:
		length = AppendString(pCommandLine, "-W", 0, length);
		nArgs += 1;
		break;
	}
	
	switch (options & kSampleFormatMask) {
	
	case kSigned8BitSamples:
		length = AppendString(pCommandLine, "-c", 0, length);
		nArgs += 1;
		break;

	case kUnsigned8BitSamples:
		length = AppendString(pCommandLine, "-8", 0, length);
		nArgs += 1;
		break;

	case kALawSamples:
		length = AppendString(pCommandLine, "-a", 0, length);
		nArgs += 1;
		break;

	case kmuLawSamples:
		length = AppendString(pCommandLine, "-u", 0, length);
		nArgs += 1;
		break;

	case kSigned16BitSamples:
		length = AppendString(pCommandLine, "-s", 0, length);
		nArgs += 1;
		break;

	case kSigned32BitSamples:
		length = AppendString(pCommandLine, "-l", 0, length);
		nArgs += 1;
		break;

	case kFloatSamples:
		length = AppendString(pCommandLine, "-f", 0, length);
		nArgs += 1;
		break;
	}

	if (options & kVerboseTranslation) {
		length = AppendString(pCommandLine, "-v", 0, length);
		nArgs += 1;
	}
	
	length = AppendString(pCommandLine, "-m", 0, length);
	char buffer[2];
	buffer[0] =		((options & kNoteAmplitudeMsgs) ? 1 : 0)
				 	+	((options & kOutOfRangeMsgs) ? 2 : 0)
				 	+	((options & kWarningMsgs) ? 4 : 0)
				 	+ '0';
	buffer[1] = '\0';
	length = AppendString(pCommandLine, buffer, 0, length);
	nArgs += 2;

	switch (options & kDisplayMask) {
	
	case kSuppressDisplays:
		length = AppendString(pCommandLine, "-d", 0, length);
		nArgs += 1;
		break;
		
	case kASCIIDisplays:
		length = AppendString(pCommandLine, "-g", 0, length);
		nArgs += 1;
		break;

	case kPostScriptDisplays:
		length = AppendString(pCommandLine, "-G", 0, length);
		nArgs += 1;
		break;
	}
	
	if (options & kRewriteSoundfileHdr) {
		length = AppendString(pCommandLine, "-R", 0, length);
		nArgs += 1;
	}
	
	if (options & kHeartbeat) {
		length = AppendString(pCommandLine, "-H2", 0, length);
		nArgs += 1;
	}
	
	if (options & kNotifyAtEndOfJob) {
		length = AppendString(pCommandLine, "-N", 0, length);
		nArgs += 1;
	}

	if (options & kTerminateAtEndOfTrk) {
		length = AppendString(pCommandLine, "-T", 0, length);
		nArgs += 1;
	}
	
	if (options & kDeferSoundfileLoads) {
		length = AppendString(pCommandLine, "-D", 0, length);
		nArgs += 1;
	}

	length = AppendString(pCommandLine, Filename(kOrchestraFile), 0, length);
	length = AppendString(pCommandLine, Filename(kScoreFile), 0, length);
	nArgs += 2;

	// Construct argument vector.
	char **const pArgv = new char *[nArgs + 1];		// accomodate terminating 0
	for (uint32 i = 0; i < nArgs; ++i) {
		pArgv[i] = pCommandLine;
		pCommandLine += strlen(pCommandLine) + 1;
	}
	pArgv[nArgs] = 0;
	
	return pArgv;
}

const char *
CsoundSynthesisJob::JobClass(void) const
{
	return kJobClass;
}

char *
CsoundSynthesisJob::CopyWindowTitle(void) const
{
	const char *pFileName = JobFilename();
	char *const pTitle =
		new char[strlen(pFileName) + sizeof(kWindowTitleFormat) + 1];
	sprintf(pTitle, kWindowTitleFormat, pFileName);
	return pTitle;
}

CsoundAbstractWindow *
CsoundSynthesisJob::CreateWindow(BPoint origin) const
{
	// Factory method: Create an appropriate window.
	char *const pTitle = CopyWindowTitle();
	CsoundAbstractWindow *pWindow =
		new CsoundSynthesisMainWindow(WindowPosition(origin), pTitle, *this);
	delete [] pTitle;
	return pWindow;
}
