//		Copyright © 1997–2000  Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//		Describes a Csound synthesis job--the normal mode of running Csound.

#ifndef CSSYNTHJOB_H
#define CSSYNTHJOB_H

#include <Message.h>
#include <Path.h>
#include <SupportDefs.h>

#include "CsAbstractJob.h"

class CsoundSynthesisJob : public CsoundAbstractJob
{
	typedef CsoundAbstractJob inherited;
public:
	
	// Filename specifiers.
	enum {
		kOrchestraFile,
		kScoreFile,
		kExtractFile,				// -x ...
		kInputFile,					// -i ...
		kLineFile,					// -L ...
		kMIDIInFile,				// -M ... or -F ...
		kOutputFile,				// -o ...
		kTranscriptFile,			// -- ...
		kMIDIOutFile,				// -Q ...
		
		kNumFilenames
	};
	
	// Numeric parameters.
	enum {
		kSoftwareSampleFrames,		// -b ...
		kHardwareSampleFrames,		// -B ...
		kSampleRateOverride,			// -r ...
		kControlRateOverride,		// -k ...
		kTempo,							// -t ...
		_Obsolete,						// was kSustainPedalThreshold,		// -P ...
		
		kNumNumbers
	};
	
	// Boolean command-line options.
	enum {
		kPreprocessScore		= 0x00000001,		// -C
//		kTranslateScore		= 0x00000002,		// -S
		kRealtimeMIDI			= 0x00000004,		// selects -M instead of -F
		kInitTimeOnly			= 0x00000008,		// -I
		kInhibitOutput			= 0x00000010,		// -n
		
		kFileFormatMask		= 0x00000060,
		kNoHeaderFormat		= 0x00000000,		// -h
		kSoundfileFormat		= 0x00000020,		// -J
		kAIFFFormat				= 0x00000040,		// -A
		kWAVFormat				= 0x00000060,		// -W
		
		kSampleFormatMask		= 0x00000380,
		kSigned8BitSamples	= 0x00000000,		// -c
		kUnsigned8BitSamples	= 0x00000080,		// -8
		kALawSamples			= 0x00000100,		// -a
		kmuLawSamples			= 0x00000180,		// -u
		kSigned16BitSamples	= 0x00000200,		// -s
		kSigned32BitSamples	= 0x00000280,		// -l
		kFloatSamples			= 0x00000300,		// -f
		
		kVerboseTranslation	= 0x00000400,		// -v
		kNoteAmplitudeMsgs	= 0x00000800,		// -m 1
		kOutOfRangeMsgs		= 0x00001000,		// -m 2
		kWarningMsgs			= 0x00002000,		// -m 4
		
		kDisplayMask			= 0x0000C000,
		kSuppressDisplays		= 0x00000000,		// -d
		kASCIIDisplays			= 0x00004000,		// -g
		kPostScriptDisplays	= 0x00008000,		// -G
		kGraphicsDisplays		= 0x0000C000,		// (none)
		
		kRewriteSoundfileHdr	= 0x00010000,		// -R
		kHeartbeat				= 0x00020000,		// -H
		kNotifyAtEndOfJob		= 0x00040000,		// -N
		kTerminateAtEndOfTrk	= 0x00080000,		// -T
		kDeferSoundfileLoads	= 0x00100000		// -D
	};

	CsoundSynthesisJob(bool assignName = true);
	CsoundSynthesisJob(const CsoundSynthesisJob &job);
	CsoundSynthesisJob &operator =(const CsoundSynthesisJob &job);	
	virtual ~CsoundSynthesisJob(void);

	virtual CsoundAbstractJob *New(void) const;
	virtual CsoundAbstractJob *Clone(void) const;
	virtual CsoundAbstractJob &Become(const CsoundAbstractJob &job);
	
	virtual bool operator ==(const CsoundAbstractJob &job) const;

	// Archiving methods.
	CsoundSynthesisJob(BMessage *pMsg);
	virtual status_t Archive(BMessage *pMsg, bool deep = true) const;
#pragma export on
	static CsoundSynthesisJob *Instantiate(BMessage *pMsg);
#pragma export reset

	// Check if the job is valid.
	virtual bool IsValid(void) const;
	
	// Return current directory for execution.
	virtual BPath CurrentDirectory(void) const;
	
	// Construct and return Csound command arguments.
	// The returned memory must be explicitly freed.
	virtual char **CopyCommandArguments(const char *prefix, int32 &nArgs) const;

	// Return a name for the class.
	virtual const char *JobClass(void) const;
	
	// Construct a title for a window.
	virtual char *CopyWindowTitle(void) const;

	// Create an instance of the concrete window class.
	virtual CsoundAbstractWindow *CreateWindow(BPoint origin) const;
};

#endif // CSSYNTHJOB_H
