//		Copyright © 1997–2000  Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//		The synthesis window for Csound.  Contains the most commonly needed
//		controls.

#include "CsSynthMainWindow.h"

#include <cstring>
#include <cmath>

#include <Alert.h>
#include <Box.h>
#include <Button.h>
#include <CheckBox.h>
#include <Font.h>
#include <MenuItem.h>
#include <Message.h>
#include <MidiEndpoint.h>
#include <MidiRoster.h>
#include <Mime.h>
#include <Node.h>
#include <NodeInfo.h>
#include <Point.h>
#include <PopUpMenu.h>
#include <Rect.h>
#include <TextControl.h>
#include <View.h>

#include "CsAbstractJob.h"
#include "CsSynthJob.h"
#include "CsAbstractMainWindow.h"
#include "CsSynthOptionsWindow.h"
#include "CsTranscriptWindow.h"
#include "MenuTextControl.h"

#define nElems(array) (sizeof(array) / sizeof(array[0]))

namespace
{
	// MIME types.
	const char kOrchestraMimeType[] = "text/x-csound-orchestra";
	const char kScoreMimeType[] = "text/x-csound-score";
	//const char kTranslatedScoreMimeType[] = "text/x-csound-scoret";
	const char kMIDIMimeType[] = "audio/midi";
	
	// Filename extensions.
	const char kOrcExtension[] = ".orc";
	const char kScoExtension[] = ".sco";
	const char kSrtExtension[] = ".srt";		// sorted score
	const char kXtrExtension[] = ".xtr";		// extract files
	const char kExtractExtension[] = ".extract";
	const char kMidiExtension[] = ".midi";	// MIDI files
	const char kMidExtension[] = ".mid";
	const char kMfExtension[] = ".mf";
	
	// Coordinates.
	const BPoint kOptionsWindowOffset(80, 60);
	
	// UI description data.
	const char kMIDIInMenuName[] = "MIDI Inputs";
	const char kMIDIOutMenuName[] = "MIDI Outputs";

	const CsoundAbstractWindow::FilenameDescriptor kFilenameDescriptors[] =
	{
		{	"orch", "Orchestra File", CsoundAbstractWindow::kMandatory,
			CsoundSynthesisJob::kOrchestraFile
		},
		{	"scor", "Score File", CsoundAbstractWindow::kMandatory,
			CsoundSynthesisJob::kScoreFile
		},
		{	"extr", "Extract File", CsoundAbstractWindow::kNormal,
			CsoundSynthesisJob::kExtractFile
		},
		{	"sinp", "Sound Input File", CsoundAbstractWindow::kNormal,
			CsoundSynthesisJob::kInputFile
		},
		{	"linp", "Line-Oriented Input", CsoundAbstractWindow::kNormal,
			CsoundSynthesisJob::kLineFile
		},
		{	"midi", "MIDI Input", CsoundAbstractWindow::kHasMenu,
			CsoundSynthesisJob::kMIDIInFile
		},
		{	"outp", "Sound Output File", CsoundAbstractWindow::kNormal,
			CsoundSynthesisJob::kOutputFile
		},
		{	"mido", "MIDI Output", CsoundAbstractWindow::kHasMenu,
			CsoundSynthesisJob::kMIDIOutFile
		},
		{	"tscr", "Transcript", CsoundAbstractWindow::kNormal,
			CsoundSynthesisJob::kTranscriptFile
		}
	};
	
	const CsoundAbstractWindow::CheckBoxDescriptor kCheckBoxDescriptors[] =
	{
	//	{	"scot", "Use SCOT Format",
	//		CsoundSynthesisJob::kTranslateScore
	//	},
		{	"rtmi", "Realtime MIDI Input",
			CsoundSynthesisJob::kRealtimeMIDI
		},
	};
	
	const CsoundAbstractWindow::ButtonDescriptor kButtonDescriptors[] =
	{
		{	"strt", "Synthesize",
			CsoundSynthesisMainWindow::kStartSynthesis
		},
		{	"abrt", "Abort",	
			CsoundSynthesisMainWindow::kAbortSynthesis
		}
	};
	
	// Callbacks for pop-up menu creation.
	class MenuCallback : public MenuTextControl::Callback
	{
		typedef MenuTextControl::Callback inherited;
		typedef bool (BMidiEndpoint::*TypeChecker)(void) const;
	public:
		MenuCallback(const char *pName, TypeChecker pTypeChecker)
		:	mpName(pName),
			mpTypeChecker(pTypeChecker)
		{
			// empty
		}
		
		virtual ~MenuCallback(void)
		{
			// empty
		}
		
		virtual BPopUpMenu *
		operator()(const BMessage &msg, BHandler *pTarget) const
		{
			BPopUpMenu *pMenu = new BPopUpMenu(mpName, false, false);
			
			// Iterate through the known MIDI endpoints, adding appropriate
			// items to the menu.
			if (BMidiRoster *pRoster = BMidiRoster::MidiRoster()) {
				int32 id = 0;
				while (BMidiEndpoint *pEndpoint = pRoster->NextEndpoint(&id)) {
					if ((pEndpoint->*mpTypeChecker)()) {
						BMenuItem *pMenuItem =
							new BMenuItem(pEndpoint->Name(), new BMessage(msg));
						pMenuItem->SetTarget(pTarget);
						pMenu->AddItem(pMenuItem);
					}
					pEndpoint->Release();
				}
			}
			
			return pMenu;
		}
	
	private:
		MenuCallback(const MenuCallback &);
		MenuCallback &operator =(const MenuCallback &);
		
		// Data members.
		const char *const mpName;
		const TypeChecker mpTypeChecker;
	};
}

CsoundSynthesisMainWindow::CsoundSynthesisMainWindow(
	BPoint origin, const char *pTitle, const CsoundSynthesisJob &job)
:	inherited(origin, pTitle, job),
	mpTranscriptWindow((CsoundTranscriptWindow *)0)
{
	// Create all the controls and determine space requirements.
	uint32 i;
	float maxWidth = 0;
	float maxDivider = 0;
	float maxButtonWidth = 0;

	CsoundAbstractWindow::ControlDimensions filenameDimensions[kNumFilenames];
	CreateControls(kNumFilenames,
						kFilenameDescriptors,
						mpFilenames,
						filenameDimensions,
						*this);
	for (i = 0; i < kNumFilenames; ++i) {
		const CsoundAbstractWindow::ControlDimensions &dim =
			filenameDimensions[i];
		if (dim.width > maxWidth) {
			maxWidth = dim.width;
		}
		if (dim.divider > maxDivider) {
			maxDivider = dim.divider;
		}
	}
	dynamic_cast<MenuTextControl &>(*mpFilenames[kMIDIInFilename]).SetCallback(
		new MenuCallback(kMIDIInMenuName, &BMidiEndpoint::IsProducer));
	dynamic_cast<MenuTextControl &>(*mpFilenames[kMIDIOutFilename]).SetCallback(
		new MenuCallback(kMIDIOutMenuName, &BMidiEndpoint::IsConsumer));

	CsoundAbstractWindow::ControlDimensions checkBoxDimensions[kNumCheckBoxes];
	CreateControls(kNumCheckBoxes,
						kCheckBoxDescriptors,
						mpCheckBoxes,
						checkBoxDimensions,
						*this);
	for (i = 0; i < kNumCheckBoxes; ++i) {
		if (checkBoxDimensions[i].width > maxWidth) {
			maxWidth = checkBoxDimensions[i].width;
		}
	}
	
	CsoundAbstractWindow::ControlDimensions buttonDimensions[kNumButtons];
	CreateControls(kNumButtons,
						kButtonDescriptors,
						mpButtons,
						buttonDimensions,
						*this);
	for (i = 0; i < kNumButtons; ++i) {
		if (buttonDimensions[i].width > maxButtonWidth) {
			maxButtonWidth = buttonDimensions[i].width;
		}
	}

	// Calculate overall space requirements and adjust sizes.
	if (maxDivider + 200 > maxWidth) {
		maxWidth = maxDivider + 200;
	}	
	if (kNumButtons * maxButtonWidth > maxWidth) {
		maxWidth = kNumButtons * maxButtonWidth;
	}
	
	for (i = 0; i < kNumFilenames; ++i) {
		BTextControl *pControl = mpFilenames[i];
		// The following had to be empirically determined :-(
		pControl->ResizeTo(maxWidth, filenameDimensions[i].height);
		pControl->SetDivider(maxDivider);
		pControl->TextView()->ResizeTo(maxWidth-maxDivider-10,
												 pControl->TextView()->Frame().Height());
	}
	for (i = 0; i < kNumButtons; ++i) {
		mpButtons[i]->ResizeTo(maxButtonWidth, buttonDimensions[i].height);
	}
	
	// Lay out the controls.
	BView *const pBackgroundView = BackgroundView();
	
	font_height fontHeight;
	be_bold_font->GetHeight(&fontHeight);
	const float spacing =
		ceil((fontHeight.ascent + fontHeight.descent + fontHeight.leading) / 2);
	
	float y = 2 * spacing;
	mpFilenames[kOrchestraFilename]->MoveTo(spacing, y);
	mpFilenames[kScoreFilename]->MoveTo(
		spacing, y += filenameDimensions[kOrchestraFilename].height);
//	mpCheckBoxes[kTranslateScore]->MoveTo(
//		spacing, y += filenameDimensions[kScoreFilename].height);
	mpFilenames[kExtractFilename]->MoveTo(
//		spacing, y += checkBoxDimensions[kTranslateScore].height);
		spacing, y += filenameDimensions[kScoreFilename].height);
	mpFilenames[kInputFilename]->MoveTo(
		spacing, y += filenameDimensions[kExtractFilename].height);
	mpFilenames[kLineFilename]->MoveTo(
		spacing, y += filenameDimensions[kInputFilename].height);
	mpFilenames[kMIDIInFilename]->MoveTo(
		spacing, y += filenameDimensions[kLineFilename].height);
	mpCheckBoxes[kRealtimeMIDI]->MoveTo(
		spacing, y += filenameDimensions[kMIDIInFilename].height);
	y += checkBoxDimensions[kRealtimeMIDI].height;
	BRect inputsFrame(B_ORIGIN, BPoint(maxWidth + 2 * spacing, y + spacing));
	inputsFrame.OffsetBy(BPoint(spacing, spacing));
	BBox *const pInputs = new BBox(inputsFrame, "inputs");
	pInputs->SetLabel("Inputs");
	pInputs->AddChild(mpFilenames[kOrchestraFilename]);
	pInputs->AddChild(mpFilenames[kScoreFilename]);
//	pInputs->AddChild(mpCheckBoxes[kTranslateScore]);
	pInputs->AddChild(mpFilenames[kExtractFilename]);
	pInputs->AddChild(mpFilenames[kInputFilename]);
	pInputs->AddChild(mpFilenames[kLineFilename]);
	pInputs->AddChild(mpFilenames[kMIDIInFilename]);
	pInputs->AddChild(mpCheckBoxes[kRealtimeMIDI]);
	pBackgroundView->AddChild(pInputs);

	y = 2 * spacing;
	mpFilenames[kOutputFilename]->MoveTo(spacing, y);
	mpFilenames[kMIDIOutFilename]->MoveTo(
		spacing, y += filenameDimensions[kOutputFilename].height);
	mpFilenames[kTranscriptFilename]->MoveTo(
		spacing, y += filenameDimensions[kMIDIOutFilename].height);
	y += filenameDimensions[kTranscriptFilename].height;
	BRect outputsFrame(B_ORIGIN, BPoint(maxWidth + 2 * spacing, y + spacing));
	outputsFrame.OffsetBy(pInputs->Frame().LeftBottom());
	outputsFrame.OffsetBy(BPoint(0, spacing));
	BBox *const pOutputs = new BBox(outputsFrame, "outputs");
	pOutputs->SetLabel("Outputs");
	pOutputs->AddChild(mpFilenames[kOutputFilename]);
	pOutputs->AddChild(mpFilenames[kMIDIOutFilename]);
	pOutputs->AddChild(mpFilenames[kTranscriptFilename]);
	pBackgroundView->AddChild(pOutputs);

	BButton *pStart = mpButtons[kStart];
	BButton *pAbort = mpButtons[kAbort];
	pStart->MoveTo(spacing, pOutputs->Frame().bottom + spacing);
	pAbort->MoveTo(pStart->Frame().RightTop() + BPoint(spacing, 0));
	pBackgroundView->AddChild(pStart);
	pBackgroundView->AddChild(pAbort);

	const float preferredWidth = maxWidth + 4 * spacing;
	const float preferredHeight = pStart->Frame().bottom + spacing;

	// Resize to final dimensions.
	pBackgroundView->ResizeTo(preferredWidth, preferredHeight);
	ResizeTo(preferredWidth, pBackgroundView->Frame().bottom);
	
	// Initialize control states.
	pStart->MakeDefault(true);
	pStart->SetEnabled(IsJobValid());
	pAbort->SetEnabled(false);

	// Create options window.
	char *const pOptionsTitle = CopyOptionsWindowTitle();
	CsoundSynthesisOptionsWindow *pOptions =
		new CsoundSynthesisOptionsWindow(
				origin + kOptionsWindowOffset, pOptionsTitle, *this);
	delete [] pOptionsTitle;
	AddOptionsWindow(pOptions, spacing, pAbort);
	pOptions->Run();
}

CsoundSynthesisMainWindow::~CsoundSynthesisMainWindow(void)
{
	// empty
}

void
CsoundSynthesisMainWindow::MessageReceived(BMessage *pMsg)
{
	switch (pMsg->what) {

	// Command button messages.
	case kStartSynthesis:
		if (!mpTranscriptWindow && IsJobValid()) {
			// Start the job in a new transcript window.
			mpTranscriptWindow =
				new CsoundTranscriptWindow(Frame().LeftTop() + BPoint(100, 100),
													 Title(), *this, "csound", CurrentJob());
			mpButtons[kStart]->SetEnabled(false);
			mpButtons[kAbort]->SetEnabled(true);			
		}
		break;
		
	case kAbortSynthesis:
		if (mpTranscriptWindow) {
			mpTranscriptWindow->PostMessage(CsoundTranscriptWindow::kAbortJob);
		}
		break;
		
	case CsoundTranscriptWindow::kJobFailed:
		if (mpTranscriptWindow && mpTranscriptWindow->Lock()) {
			mpTranscriptWindow->Quit();
		}
		(new BAlert("", "Failed to start synthesis.", "Duh!"))->Go();
		// FALL THROUGH
	case CsoundTranscriptWindow::kJobFinished:
		mpTranscriptWindow = (CsoundTranscriptWindow *)0;	
		mpButtons[kStart]->SetEnabled(IsJobValid());
		mpButtons[kAbort]->SetEnabled(false);
		break;

	// Fall back.
	default:
		inherited::MessageReceived(pMsg);
		break;
	}
}

void
CsoundSynthesisMainWindow::UpdateControlValues(void)
{
	// Update all controls to their new values.
	SetControlValues(kNumFilenames, kFilenameDescriptors, mpFilenames, *this);
	SetControlValues(kNumCheckBoxes, kCheckBoxDescriptors, mpCheckBoxes, *this);
	
	inherited::UpdateControlValues();
}

void
CsoundSynthesisMainWindow::ValidJob(bool valid)
{
	if (!mpTranscriptWindow) {
		mpButtons[kStart]->SetEnabled(valid);
	}
}

void
CsoundSynthesisMainWindow::AcceptDroppedRef(BPoint where, const entry_ref *pRef)
{
	BPath path;
	BEntry entry(pRef);
	if (entry.GetPath(&path) == B_NO_ERROR) {
		const char *pPath = path.Path();
		BTextControl *pControl = (BTextControl *)0;
		
		// First try to map the file's mime type to one of our controls.
		static struct {
			const char *pMimeType;
			const uint32 filenameIndex;
		} mimeMaps[] = {
			{ kOrchestraMimeType, kOrchestraFilename },
			{ kScoreMimeType, kScoreFilename },
//			{ kTranslatedScoreMimeType, kScoreFilename },
			{ kMIDIMimeType, kMIDIInFilename }
		};
		const int nMimeMaps = nElems(mimeMaps);
	
		BNode node(pRef);
		BNodeInfo nodeInfo(&node);
		char mimeType[B_MIME_TYPE_LENGTH];
		if (nodeInfo.GetType(mimeType) == B_NO_ERROR) {
			for (int i = 0; !pControl && i < nMimeMaps; ++i) {
				if (strcmp(mimeType, mimeMaps[i].pMimeType) == 0) {
					pControl = mpFilenames[mimeMaps[i].filenameIndex];
				}
			}
		}
		
		// If that failed, try to map the file's extension to a control.
		// ### On success, set the mime type.
		static struct {
			const char *pExtension;
			const char *pMimeType;
			const uint32 filenameIndex;
		} extensionMaps[] = {
			{ kOrcExtension, kOrchestraMimeType, kOrchestraFilename },
			{ kScoExtension, kScoreMimeType, kScoreFilename },
			{ kSrtExtension, kScoreMimeType, kScoreFilename },
			{ kXtrExtension, (const char *)0, kExtractFilename },
			{ kExtractExtension, (const char *)0, kExtractFilename },
			{ kMidiExtension, kMIDIMimeType, kMIDIInFilename },
			{ kMidExtension, kMIDIMimeType, kMIDIInFilename },
			{ kMfExtension, kMIDIMimeType, kMIDIInFilename }
		};
		const int nExtensionMaps = nElems(extensionMaps);

		const char *pExtension = strrchr(pPath, '.');
		if (pExtension) {
			for (int i = 0; !pControl && i < nExtensionMaps; ++i) {
				if (strcmp(pExtension, extensionMaps[i].pExtension) == 0) {
					pControl = mpFilenames[extensionMaps[i].filenameIndex];
					
					const char *pMimeType = extensionMaps[i].pMimeType;
					if (pMimeType) {
						nodeInfo.SetType(pMimeType);
					}
				}
			}
		}
		
		if (pControl) {
			// OK, we found a suitable control.
			pControl->SetText(path.Path());
			pControl->Invoke();
			return;
		}
	}
	
	// Fall back to inherited behavior.
	inherited::AcceptDroppedRef(where, pRef);
}
