//		Copyright © 1997–2000  Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//		The synthesis options window; contains less frequently used settings.

#include <cmath>

#include <Box.h>
#include <CheckBox.h>
#include <Font.h>
#include <Point.h>
#include <RadioButton.h>
#include <Rect.h>

#include "CsSynthJob.h"
#include "CsSynthMainWindow.h"
#include "CsSynthOptionsWindow.h"

// UI description data.
const CsoundAbstractWindow::CheckBoxDescriptor kCheckBoxDescriptors[] =
{
	{	"vbxl", "Verbose Orchestra Translation",
		CsoundSynthesisJob::kVerboseTranslation
	},
	{	"ampm", "Display Note Amplitude Messages",
		CsoundSynthesisJob::kNoteAmplitudeMsgs
	},
	{	"oorm", "Display Out-of-Range Messages",
		CsoundSynthesisJob::kOutOfRangeMsgs
	},
	{	"warn", "Display Warnings",
		CsoundSynthesisJob::kWarningMsgs
	},
	{	"htbt", "Show Heartbeat",
		CsoundSynthesisJob::kHeartbeat
	},
	{	"neoj", "Notify at End of Job",
		CsoundSynthesisJob::kNotifyAtEndOfJob
	},
	{	"ppsc", "Preprocess Score",
		CsoundSynthesisJob::kPreprocessScore
	},
	{	"iton", "Init-Time Processing Only",
		CsoundSynthesisJob::kInitTimeOnly
	},
	{	"inho", "Inhibit Soundfile Output",
		CsoundSynthesisJob::kInhibitOutput
	},
	{	"rsfh", "Rewrite Soundfile Header",
		CsoundSynthesisJob::kRewriteSoundfileHdr
	},
	{	"dsfl", "Defer Soundfile Loading",
		CsoundSynthesisJob::kDeferSoundfileLoads
	},
	{	"teot", "Terminate at End of MIDI Track",
		CsoundSynthesisJob::kTerminateAtEndOfTrk
	}
};

const CsoundAbstractWindow::NumberDescriptor kNumberDescriptors[] =
{
	{	"swsf", "Software Buffer Size", "%.0f",
		CsoundSynthesisJob::kSoftwareSampleFrames
	},
	{	"hwsf", "Hardware Buffer Size", "%.0f",
		CsoundSynthesisJob::kHardwareSampleFrames
	},
	{	"sovr", "Sample Rate Override", "%.0f",
		CsoundSynthesisJob::kSampleRateOverride
	},
	{	"kovr", "Control Rate Override", "%.0f",
		CsoundSynthesisJob::kControlRateOverride
	},
	{	"tmpo", "Tempo", "%.0f",
		CsoundSynthesisJob::kTempo
	}
//	{	"spth", "Sustain Pedal Threshold", "%.0f",
//		CsoundSynthesisJob::kSustainPedalThreshold
//	}
};

const CsoundAbstractWindow::RadioButtonDescriptor kRadioButtonDescriptors[] =
{
	{	"sgn8", "8-Bit Signed",
		CsoundSynthesisJob::kSampleFormatMask,
		CsoundSynthesisJob::kSigned8BitSamples
	},
	{	"uns8", "8-Bit Unsigned",
		CsoundSynthesisJob::kSampleFormatMask,
		CsoundSynthesisJob::kUnsigned8BitSamples
	},
	{	"alaw", "A-Law",
		CsoundSynthesisJob::kSampleFormatMask,
		CsoundSynthesisJob::kALawSamples
	},
	{	"ulaw", "µ-Law",
		CsoundSynthesisJob::kSampleFormatMask,
		CsoundSynthesisJob::kmuLawSamples
	},
	{	"sg16", "16-Bit Signed",
		CsoundSynthesisJob::kSampleFormatMask,
		CsoundSynthesisJob::kSigned16BitSamples
	},
	{	"sg32", "32-Bit Signed",
		CsoundSynthesisJob::kSampleFormatMask,
		CsoundSynthesisJob::kSigned32BitSamples
	},
	{	"flot", "32-Bit Float",
		CsoundSynthesisJob::kSampleFormatMask,
		CsoundSynthesisJob::kFloatSamples
	},
	{	"nhdr", "Raw Data",
		CsoundSynthesisJob::kFileFormatMask,
		CsoundSynthesisJob::kNoHeaderFormat
	},
	{	"sndf", "IRCAM",
		CsoundSynthesisJob::kFileFormatMask,
		CsoundSynthesisJob::kSoundfileFormat
	},
	{	"aiff", "AIFF",
		CsoundSynthesisJob::kFileFormatMask,
		CsoundSynthesisJob::kAIFFFormat
	},
	{	"wave", "WAV",
		CsoundSynthesisJob::kFileFormatMask,
		CsoundSynthesisJob::kWAVFormat
	},
	{	"sdpy", "None",
		CsoundSynthesisJob::kDisplayMask,
		CsoundSynthesisJob::kSuppressDisplays
	},
	{	"adpy", "ASCII",
		CsoundSynthesisJob::kDisplayMask,
		CsoundSynthesisJob::kASCIIDisplays
	},
	{	"pdpy", "PostScript",
		CsoundSynthesisJob::kDisplayMask,
		CsoundSynthesisJob::kPostScriptDisplays
	},
	{	"gdpy", "Graphics",
		CsoundSynthesisJob::kDisplayMask,
		CsoundSynthesisJob::kGraphicsDisplays
	}
};

CsoundSynthesisOptionsWindow::CsoundSynthesisOptionsWindow(
	BPoint origin, const char *pTitle, CsoundSynthesisMainWindow &target)
:	inherited(origin, pTitle, target)
{
	// Create all the controls and determine space requirements.
	uint32 i;
	float maxWidth = 0;
	float maxDivider = 0;
	float maxCheckBoxWidth = 0;
	float maxRadioButtonWidth = 0;

	CsoundAbstractWindow::ControlDimensions checkBoxDimensions[kNumCheckBoxes];
	CreateControls(kNumCheckBoxes,
						kCheckBoxDescriptors,
						mpCheckBoxes,
						checkBoxDimensions,
						target);
	for (i = 0; i < kNumCheckBoxes; ++i) {
		const CsoundAbstractWindow::ControlDimensions &dim =
			checkBoxDimensions[i];
		if (dim.width > maxWidth) {
			maxWidth = dim.width;
		}
		if (dim.width > maxCheckBoxWidth) {
			maxCheckBoxWidth = dim.width;
		}
	}

	CsoundAbstractWindow::ControlDimensions numberDimensions[kNumNumbers];
	CreateControls(kNumNumbers,
						kNumberDescriptors,
						mpNumbers,
						numberDimensions,
						target);
	for (i = 0; i < kNumNumbers; ++i) {
		const CsoundAbstractWindow::ControlDimensions &dim =
			numberDimensions[i];
		if (dim.width > maxWidth) {
			maxWidth = dim.width;
		}
		if (dim.divider > maxDivider) {
			maxDivider = dim.divider;
		}
	}

	CsoundAbstractWindow::ControlDimensions radioButtonDimensions[kNumRadioButtons];
	CreateControls(kNumRadioButtons,
						kRadioButtonDescriptors,
						mpRadioButtons,
						radioButtonDimensions,
						target);
	for (i = 0; i < kNumRadioButtons; ++i) {
		const CsoundAbstractWindow::ControlDimensions &dim =
			radioButtonDimensions[i];
		if (dim.width > maxWidth) {
			maxWidth = dim.width;
		}
		if (dim.width > maxRadioButtonWidth) {
			maxRadioButtonWidth = dim.width;
		}
	}

	// Calculate overall space requirements and adjust sizes.
	font_height fontHeight;
	be_bold_font->GetHeight(&fontHeight);
	const float spacing =
		ceil((fontHeight.ascent + fontHeight.descent + fontHeight.leading) / 2);
	
	if (2 * maxCheckBoxWidth > maxWidth) {
		maxWidth = 2 * maxCheckBoxWidth;
	}
	if (maxDivider + 200 > maxWidth) {
		maxWidth = maxDivider + 200;
	}	
	if (3 * maxRadioButtonWidth + 6 * spacing > maxWidth) {
		maxWidth = 3 * maxRadioButtonWidth + 6 * spacing;
	}

	for (i = 0; i < kNumNumbers; ++i) {
		BTextControl *pControl = mpNumbers[i];
		// The following had to be empirically determined :-(
		pControl->ResizeTo(maxWidth, numberDimensions[i].height);
		pControl->SetDivider(maxDivider);
		pControl->TextView()->ResizeTo(maxWidth-maxDivider-10,
												 pControl->TextView()->Frame().Height());
	}

	// Lay out the controls.
	BView *const pBackgroundView = BackgroundView();
	
	float y = 2 * spacing;
	for (i = 0; i < (kNumCheckBoxes+1)/2; ++i) {
		mpCheckBoxes[i]->MoveTo(spacing, y);
		y += checkBoxDimensions[i].height;
	}
	float y1 = 2 * spacing;
	while (i < kNumCheckBoxes) {
		mpCheckBoxes[i]->MoveTo(spacing + maxWidth/2, y1);
		y1 += checkBoxDimensions[i].height;
		++i;
	}
	BRect switchesFrame(B_ORIGIN,
							  BPoint(maxWidth + 2 * spacing, y + spacing));
	switchesFrame.OffsetBy(BPoint(spacing, spacing));
	BBox *const pSwitches = new BBox(switchesFrame, "switches");
	pSwitches->SetLabel("Switches");
	for (i = 0; i < kNumCheckBoxes; ++i) {
		pSwitches->AddChild(mpCheckBoxes[i]);
	}
	pBackgroundView->AddChild(pSwitches);

	y = 2 * spacing;
	for (i = 0; i < kNumNumbers; ++i) {
		mpNumbers[i]->MoveTo(spacing, y);
		y += numberDimensions[i].height;
	}
	BRect numbersFrame(B_ORIGIN, BPoint(maxWidth + 2 * spacing, y + spacing));
	numbersFrame.OffsetBy(pSwitches->Frame().LeftBottom());
	numbersFrame.OffsetBy(BPoint(0, spacing));
	BBox *const pNumbers = new BBox(numbersFrame, "numbers");
	pNumbers->SetLabel("Numeric Parameters");
	for (i = 0; i < kNumNumbers; ++i) {
		pNumbers->AddChild(mpNumbers[i]);
	}
	pBackgroundView->AddChild(pNumbers);

	y = 2 * spacing;
	for (i = kSigned8BitSamples; i <= kFloatSamples; ++i) {
		mpRadioButtons[i]->MoveTo(spacing, y);
		y += radioButtonDimensions[i].height;
	}
	BRect sampleFormatsFrame(B_ORIGIN, BPoint(maxWidth/3, y + spacing));
	sampleFormatsFrame.OffsetBy(pNumbers->Frame().LeftBottom());
	sampleFormatsFrame.OffsetBy(BPoint(0, spacing));
	BBox *const pSampleFormats = new BBox(sampleFormatsFrame, "sampleFormats");
	pSampleFormats->SetLabel("Sample Formats");
	for (i = kSigned8BitSamples; i <= kFloatSamples; ++i) {
		pSampleFormats->AddChild(mpRadioButtons[i]);
	}
	pBackgroundView->AddChild(pSampleFormats);

	y1 = 2 * spacing;
	for (i = kNoHeaderFormat; i <= kWAVFormat; ++i) {
		mpRadioButtons[i]->MoveTo(spacing, y1);
		y1 += radioButtonDimensions[i].height;
	}
	BRect fileFormatsFrame(B_ORIGIN, BPoint(maxWidth/3, y + spacing));
	fileFormatsFrame.OffsetBy(pSampleFormats->Frame().RightTop());
	fileFormatsFrame.OffsetBy(BPoint(spacing, 0));
	BBox *const pFileFormats = new BBox(fileFormatsFrame, "fileFormats");
	pFileFormats->SetLabel("File Formats");
	for (i = kNoHeaderFormat; i <= kWAVFormat; ++i) {
		pFileFormats->AddChild(mpRadioButtons[i]);
	}
	pBackgroundView->AddChild(pFileFormats);

	y1 = 2 * spacing;
	for (i = kSuppressDisplays; i <= kGraphicsDisplays; ++i) {
		mpRadioButtons[i]->MoveTo(spacing, y1);
		y1 += radioButtonDimensions[i].height;
	}
	BRect displaysFrame(B_ORIGIN, BPoint(maxWidth/3, y + spacing));
	displaysFrame.OffsetBy(pFileFormats->Frame().RightTop());
	displaysFrame.OffsetBy(BPoint(spacing, 0));
	displaysFrame.right = maxWidth + 3*spacing;	// compensate for rounding errors
	BBox *const pDisplays = new BBox(displaysFrame, "displays");
	pDisplays->SetLabel("Displays");
	for (i = kSuppressDisplays; i <= kGraphicsDisplays; ++i) {
		pDisplays->AddChild(mpRadioButtons[i]);
	}
	pBackgroundView->AddChild(pDisplays);

	// Resize to final dimensions.
	const float preferredWidth = maxWidth + 4 * spacing;
	const float preferredHeight = pSampleFormats->Frame().bottom + spacing;
	pBackgroundView->ResizeTo(preferredWidth, preferredHeight);
	ResizeTo(preferredWidth, pBackgroundView->Frame().bottom);
}

CsoundSynthesisOptionsWindow::~CsoundSynthesisOptionsWindow(void)
{
	// empty
}

void
CsoundSynthesisOptionsWindow::UpdateControlValues(void)
{
	// Update all controls to their new values.
	SetControlValues(kNumCheckBoxes,
						  kCheckBoxDescriptors,
						  mpCheckBoxes,
						  Target());
	SetControlValues(kNumNumbers,
						  kNumberDescriptors,
						  mpNumbers,
						  Target());
	SetControlValues(kNumRadioButtons,
						  kRadioButtonDescriptors,
						  mpRadioButtons,
						  Target());
}
