//		Copyright © 1999–2000  Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//		A text control with associated menu of predefined strings.

#include "MenuTextControl.h"

#include <MenuItem.h>
class BMessage;
#include <Point.h>
#include <PopUpMenu.h>
#include <Rect.h>
#include <TextControl.h>

namespace
{
	// Message fields.
	const char kFieldSource[] = "source";

	// Size of arrow in menu 'button'.
	const float kMenuArrowSize = 6.0;

	// Colors.
	const rgb_color kWhite = { 255, 255, 255, 255 }; 
	const rgb_color kLightGray = { 152, 152, 152, 255 }; 
	const rgb_color kDarkGray = { 100, 100, 100, 255 }; 
	const rgb_color kBlack = { 0, 0, 0, 255};
}

MenuTextControl::MenuTextControl(
	BRect frame,
	const char *pName, const char *pLabel, const char *pInitialText, 
	BMessage *pMessage, uint32 rmask, uint32 flags)
:	inherited(frame, pName, pLabel, pInitialText, pMessage, rmask, flags),
	mpCallback(0)
{
	// empty
}

MenuTextControl::~MenuTextControl(void)
{
	delete mpCallback;
	mpCallback = 0;
}
	
void
MenuTextControl::Draw(BRect updateRect)
{
	inherited::Draw(updateRect);

	// Draw the menu 'button'.	
	const rgb_color white = AdjustColor(kWhite);
	const rgb_color lightGray = AdjustColor(kLightGray);
	const rgb_color darkGray = AdjustColor(kDarkGray);
	const rgb_color black = AdjustColor(kBlack);

	const BRect frame = ArrowFrame();
	BeginLineArray(10);
	AddLine(frame.LeftBottom(), frame.LeftTop(), lightGray);
	AddLine(frame.LeftTop(), frame.RightTop(), lightGray);
	AddLine(frame.LeftBottom() + BPoint(1.0, -1.0),
			  frame.LeftTop() + BPoint(1.0, 1.0),
			  white);
	AddLine(frame.LeftTop() + BPoint(1.0, 1.0),
			  frame.RightTop() + BPoint(-1.0, 1.0),
			  white);
	AddLine(frame.LeftBottom() + BPoint(1.0, 0.0),
			  frame.RightBottom(),
			  darkGray);
	AddLine(frame.RightBottom(),
			  frame.RightTop() + BPoint(0.0, 1.0),
			  darkGray);
	AddLine(frame.LeftBottom() + BPoint(2.0, -1.0),
			  frame.RightBottom() + BPoint(-1.0, -1.0),
			  lightGray);
	AddLine(frame.RightBottom() + BPoint(-1.0, -1.0),
			  frame.RightTop() + BPoint(-1.0, 2.0),
			  lightGray);
	AddLine(frame.LeftBottom() + BPoint(1.0, 1.0),
			  frame.RightBottom() + BPoint(1.0, 1.0),
			  lightGray);
	AddLine(frame.RightBottom() + BPoint(1.0, 1.0),
			  frame.RightTop() + BPoint(1.0, 1.0),
			  lightGray);
	EndLineArray();

	const float height = frame.Height();
	const float y0 = kMenuArrowSize;
	const float x0 = y0 * sqrt(3.0) * 0.5;
	const float y1 = (height - y0) * 0.5;
	const float x1 = (height - x0) * 0.5;
	const BPoint p = frame.LeftTop();
	const BPoint points[3] = {
		p + BPoint(x1, y1),
		p + BPoint(x1 + x0, y1 + y0 * 0.5),
		p + BPoint(x1, y1 + y0)
	};
	const BRect polyFrame(p.x + x1, p.y + y1, p.x + x1 + x0, p.y + y1 + y0);
	
	const rgb_color color = HighColor();
	SetHighColor(lightGray);
	FillPolygon(points, 3, polyFrame);
	SetHighColor(black);
	StrokePolygon(points, 3, polyFrame);
	SetHighColor(color);
}

void
MenuTextControl::MessageReceived(BMessage *pMsg)
{
	switch (pMsg->what) {
	
	case kReplaceValue:
		{	void *pPointer;
			if (pMsg->FindPointer(kFieldSource, &pPointer) == B_OK && pPointer) {
				BMenuItem *pSource = static_cast<BMenuItem *>(pPointer);
				SetText(pSource->Label());
				Invoke();
			}
		}
		break;
	
	default:
		inherited::MessageReceived(pMsg);
		break;
	}
}
	
void
MenuTextControl::MouseDown(BPoint where)
{
	inherited::MouseDown(where);
	
	if (IsEnabled() && ArrowFrame().Contains(where) && mpCallback) {
		// Create and pop up a menu for replacing the contents.
		BMessage msg(kReplaceValue);
		if (BPopUpMenu *pMenu = (*mpCallback)(msg, this)) {
			const BPoint point =
				ConvertToScreen(ChildAt(0)->Frame().LeftBottom() + BPoint(-2, 2));
			pMenu->SetAsyncAutoDestruct(true);
			pMenu->Go(point, true, false, true);
		}
	}
}

void
MenuTextControl::SetCallback(Callback *pCallback)
{
	delete mpCallback;
	mpCallback = pCallback;
}

BRect
MenuTextControl::ArrowFrame(void)
{
	BRect frame = Bounds();
	frame.right = Divider();
	frame.left = frame.right - frame.Height();
	return frame.InsetBySelf(2.0, 2.0).OffsetBySelf(2.0, -1.0);
}

rgb_color
MenuTextControl::AdjustColor(rgb_color color) const
{
	if (!IsEnabled()) {
		return tint_color(color, B_LIGHTEN_1_TINT);
	}
	return color;
}
